"""Extra utilities. Used for example scripts."""

from ._record3d import Record3dFrame as Record3dFrame
from ._record3d import Record3dLoader as Record3dLoader
from ._urdf import ViserUrdf as ViserUrdf



"""Clean uPlot TypedDict definitions."""

# =========================================================================
# WARNING: This file is auto-generated by _generate_types.py
# DO NOT EDIT MANUALLY - Your changes will be overwritten!
# To modify types, edit the generation script or the TypeScript definitions
# =========================================================================

from __future__ import annotations

from typing import Any, Dict, Literal, Tuple, Union

from typing_extensions import Never, Required, TypedDict

# Semantic type aliases for unsupported/complex TypeScript patterns
JSCallback = Never  # JavaScript function signatures
DOMElement = Never  # DOM elements (HTMLElement, etc.)
CSSValue = str  # CSS property values
UnknownType = Any  # Unknown interface references

# Type aliases for index signatures
# Index signature: key -> Scale.
Scales = Dict[str, "Scale"]

Legend_Markers = TypedDict(
    "Legend_Markers",
    {
        "show": bool,
        # series indicator line width.
        "width": Union[float, JSCallback],
        # series indicator stroke (CSS borderColor).
        "stroke": CSSValue,
        # series indicator fill.
        "fill": CSSValue,
        # series indicator stroke style (CSS borderStyle).
        "dash": CSSValue,
    },
    total=False,
)
""":py:class:`~typing.TypedDict` type for uPlot."""

Focus = TypedDict(
    "Focus",
    {
        # alpha-transparancy of de-focused series.
        "alpha": float
    },
)
""":py:class:`~typing.TypedDict` type for uPlot."""

BBox = TypedDict(
    "BBox",
    {
        "left": Required[float],
        "top": Required[float],
        "width": Required[float],
        "height": Required[float],
        "show": bool,
    },
    total=False,
)
""":py:class:`~typing.TypedDict` type for uPlot."""

Select = TypedDict(
    "Select",
    {
        "left": Required[float],
        "top": Required[float],
        "width": Required[float],
        "height": Required[float],
        "show": bool,
        # div into which .u-select will be placed: .u-over or .u-under.
        "over": bool,
    },
    total=False,
)
""":py:class:`~typing.TypedDict` type for uPlot."""

Cursor_Bind = TypedDict(
    "Cursor_Bind",
    {
        "mousedown": JSCallback,
        "mouseup": JSCallback,
        "click": JSCallback,
        "dblclick": JSCallback,
        "mousemove": JSCallback,
        "mouseleave": JSCallback,
        "mouseenter": JSCallback,
    },
    total=False,
)
""":py:class:`~typing.TypedDict` type for uPlot."""

Cursor_Drag = TypedDict(
    "Cursor_Drag",
    {
        "setScale": bool,
        # toggles dragging along x.
        "x": bool,
        # toggles dragging along y.
        "y": bool,
        # min drag distance threshold.
        "dist": float,
        # when x & y are true, sets an upper drag limit in CSS px for adaptive/unidirectional behavior.
        "uni": float,
        # post-drag "click" event proxy, default is to prevent these click events.
        "click": UnknownType,
    },
    total=False,
)
""":py:class:`~typing.TypedDict` type for uPlot."""

Cursor_Hover = TypedDict(
    "Cursor_Hover",
    {
        # minimum cursor proximity to datapoint in CSS pixels for point hover.
        "prox": Union[float, None, UnknownType],
        # when non-zero, will only proximity-test indices forward or backward.
        "bias": Literal[0, 1, -1],
        # what values to treat as non-hoverable and trigger scanning to another index.
        "skip": Tuple[Any, ...],
    },
    total=False,
)
""":py:class:`~typing.TypedDict` type for uPlot."""

Cursor_Focus = TypedDict(
    "Cursor_Focus",
    {
        # minimum cursor proximity to datapoint in CSS pixels for focus activation, disabled: < 0, enabled: <= 1e6.
        "prox": Required[float],
        # when non-zero, will only focus next series towards or away from zero.
        "bias": Literal[0, 1, -1],
        # measures cursor y distance to a series in CSS pixels (for triggering setSeries hook with closest).
        "dist": UnknownType,
    },
    total=False,
)
""":py:class:`~typing.TypedDict` type for uPlot."""

Scale = TypedDict(
    "Scale",
    {
        # is this scale temporal, with series' data in UNIX timestamps?
        "time": bool,
        # determines whether all series' data on this scale will be scanned to find the full min/max range.
        "auto": Union[bool, JSCallback],
        # can define a static scale range or re-range an initially-determined range from series data.
        "range": Union[
            Tuple[Union[float, None], Union[float, None]], JSCallback, UnknownType
        ],
        # scale key from which this scale is derived.
        "from": str,
        # scale distribution. 1: linear, 2: ordinal, 3: logarithmic, 4: arcsinh.
        "distr": Literal[1, 2, 3, 4, 100],
        # logarithmic base.
        "log": Union[Literal[10], Literal[2]],
        # clamps log scale values <= 0 (default = scaleMin / 10).
        "clamp": Union[float, JSCallback],
        # arcsinh linear threshold.
        "asinh": float,
        # forward transform fn, with custom distr: 100.
        "fwd": JSCallback,
        # backward transform fn, with custom distr: 100.
        "bwd": JSCallback,
        # current min scale value.
        "min": float,
        # current max scale value.
        "max": float,
        # scale direction.
        "dir": Union[Literal[1], Literal[-1]],
        # scale orientation - 0: hz, 1: vt.
        "ori": Union[Literal[0], Literal[1]],
        # own key (for read-back).
        "key": str,
    },
    total=False,
)
""":py:class:`~typing.TypedDict` type for uPlot."""

Series_Points = TypedDict(
    "Series_Points",
    {
        # if boolean or returns boolean, round points are drawn with defined options, else fn should draw own custom points via self.ctx.
        "show": Union[bool, JSCallback],
        "paths": JSCallback,
        # may return an array of points indices to draw.
        "filter": Union[Tuple[float, ...], None, JSCallback],
        # diameter of point in CSS pixels.
        "size": float,
        # minimum avg space between point centers before they're shown (default: size * 2).
        "space": float,
        # line width of circle outline in CSS pixels.
        "width": float,
        # line color of circle outline (defaults to series.stroke).
        "stroke": CSSValue,
        # line dash segment array.
        "dash": Tuple[float, ...],
        # line cap.
        "cap": CSSValue,
        # fill color of circle (defaults to #fff).
        "fill": CSSValue,
    },
    total=False,
)
""":py:class:`~typing.TypedDict` type for uPlot."""

Series_Facet = TypedDict(
    "Series_Facet",
    {"scale": Required[str], "auto": bool, "sorted": Literal[0, 1, -1]},
    total=False,
)
""":py:class:`~typing.TypedDict` type for uPlot."""

Band = TypedDict(
    "Band",
    {
        # series indices of upper and lower band edges.
        "series": Required[Tuple[float, float]],
        # area fill style.
        "fill": CSSValue,
        # whether to fill towards yMin (-1) or yMax (+1) between "from" & "to" series.
        "dir": Union[Literal[1], Literal[-1]],
    },
    total=False,
)
""":py:class:`~typing.TypedDict` type for uPlot."""

Axis_OrthoLines = TypedDict(
    "Axis_OrthoLines",
    {
        # on/off.
        "show": bool,
        # line color.
        "stroke": CSSValue,
        # line width in CSS pixels.
        "width": float,
        # line dash segment array.
        "dash": Tuple[float, ...],
        # line cap.
        "cap": CSSValue,
    },
    total=False,
)
""":py:class:`~typing.TypedDict` type for uPlot."""

Axis_Border = TypedDict(
    "Axis_Border",
    {
        # on/off.
        "show": bool,
        # line color.
        "stroke": CSSValue,
        # line width in CSS pixels.
        "width": float,
        # line dash segment array.
        "dash": Tuple[float, ...],
        # line cap.
        "cap": CSSValue,
    },
    total=False,
)
""":py:class:`~typing.TypedDict` type for uPlot."""

Axis_FilterableOrthoLines = TypedDict(
    "Axis_FilterableOrthoLines",
    {
        # on/off.
        "show": bool,
        # line color.
        "stroke": CSSValue,
        # line width in CSS pixels.
        "width": float,
        # line dash segment array.
        "dash": Tuple[float, ...],
        # line cap.
        "cap": CSSValue,
        # can filter which splits render lines. e.g splits.map(v => v % 2 == 0 ? v : null).
        "filter": JSCallback,
    },
    total=False,
)
""":py:class:`~typing.TypedDict` type for uPlot."""

Axis_Grid = TypedDict(
    "Axis_Grid",
    {
        # on/off.
        "show": bool,
        # line color.
        "stroke": CSSValue,
        # line width in CSS pixels.
        "width": float,
        # line dash segment array.
        "dash": Tuple[float, ...],
        # line cap.
        "cap": CSSValue,
        # can filter which splits render lines. e.g splits.map(v => v % 2 == 0 ? v : null).
        "filter": JSCallback,
    },
    total=False,
)
""":py:class:`~typing.TypedDict` type for uPlot."""

Axis_Ticks = TypedDict(
    "Axis_Ticks",
    {
        # on/off.
        "show": bool,
        # line color.
        "stroke": CSSValue,
        # line width in CSS pixels.
        "width": float,
        # line dash segment array.
        "dash": Tuple[float, ...],
        # line cap.
        "cap": CSSValue,
        # can filter which splits render lines. e.g splits.map(v => v % 2 == 0 ? v : null).
        "filter": JSCallback,
        # length of tick in CSS pixels.
        "size": float,
    },
    total=False,
)
""":py:class:`~typing.TypedDict` type for uPlot."""

Plugin = TypedDict(
    "Plugin",
    {
        "hooks": Required[UnknownType],
        # can mutate provided opts as necessary.
        "opts": UnknownType,
    },
    total=False,
)
""":py:class:`~typing.TypedDict` type for uPlot."""

Cursor_Sync = TypedDict(
    "Cursor_Sync",
    {
        # sync key must match between all charts in a synced group.
        "key": Required[str],
        # determines if series toggling and focus via cursor is synced across charts.
        "setSeries": bool,
        # sets the x and y scales to sync by values. null will sync by relative (%) position.
        "scales": Tuple[Union[str, None], Union[str, None]],
        # fns that match x and y scale keys and seriesIdxs between publisher and subscriber.
        "match": Tuple[JSCallback, JSCallback, UnknownType, UnknownType, JSCallback],
        # event filters.
        "filters": UnknownType,
        # sync scales' values at the cursor position (exposed for read-back by subscribers).
        "values": Tuple[float, float],
    },
    total=False,
)
""":py:class:`~typing.TypedDict` type for uPlot."""

Legend = TypedDict(
    "Legend",
    {
        "show": bool,
        # show series values at current cursor.idx.
        "live": bool,
        # switches primary interaction mode to toggle-one/toggle-all.
        "isolate": bool,
        # series indicators.
        "markers": Legend_Markers,
        # callback for moving the legend elsewhere. e.g. external DOM container.
        "mount": UnknownType,
        # current index (readback-only, not for init).
        "idx": Union[float, None],
        # current indices (readback-only, not for init).
        "idxs": Tuple[Union[float, None], ...],
        # current values (readback-only, not for init).
        "values": Tuple[Union[str, JSCallback], ...],
    },
    total=False,
)
""":py:class:`~typing.TypedDict` type for uPlot."""

Cursor_Points = TypedDict(
    "Cursor_Points",
    {
        "show": Union[bool, JSCallback],
        # only show single y-closest point on hover (only works when cursor.focus.prox >= 0).
        "one": bool,
        # hover point diameter in CSS pixels.
        "size": Union[float, JSCallback],
        # hover point bbox in CSS pixels (will be used instead of size).
        "bbox": JSCallback,
        # hover point outline width in CSS pixels.
        "width": Union[float, JSCallback],
        # hover point outline color, pattern or gradient.
        "stroke": CSSValue,
        # hover point fill color, pattern or gradient.
        "fill": CSSValue,
    },
    total=False,
)
""":py:class:`~typing.TypedDict` type for uPlot."""

Series = TypedDict(
    "Series",
    {
        # series on/off. when off, it will not affect its scale.
        "show": bool,
        # className to add to legend parts and cursor hover points.
        "class": str,
        # scale key.
        "scale": str,
        # whether this series' data is scanned during auto-ranging of its scale.
        "auto": bool,
        # if & how the data is pre-sorted (scale.auto optimization).
        "sorted": Literal[0, 1, -1],
        # when true, null data values will not cause line breaks.
        "spanGaps": bool,
        # may mutate and/or augment gaps array found from null values.
        "gaps": Union[Tuple[Tuple[float, float], ...], JSCallback],
        # whether path and point drawing should offset canvas to try drawing crisp lines.
        "pxAlign": Union[float, bool],
        # legend label.
        "label": Union[str, DOMElement],
        # inline-legend value formatter. can be an fmtDate formatting string when scale.time: true.
        "value": Union[str, JSCallback],
        # table-legend multi-values formatter.
        "values": JSCallback,
        "paths": JSCallback,
        # rendered datapoints.
        "points": Series_Points,
        # facets.
        "facets": Tuple[Series_Facet, ...],
        # line width in CSS pixels.
        "width": float,
        # line & legend color.
        "stroke": CSSValue,
        # area fill & legend color.
        "fill": CSSValue,
        # area fill baseline (default: 0).
        "fillTo": Union[float, JSCallback],
        # line dash segment array.
        "dash": Tuple[float, ...],
        # line cap.
        "cap": CSSValue,
        # alpha-transparancy.
        "alpha": float,
        # current min and max data indices rendered.
        "idxs": Tuple[float, float],
        # current min rendered value.
        "min": float,
        # current max rendered value.
        "max": float,
    },
    total=False,
)
""":py:class:`~typing.TypedDict` type for uPlot."""

Axis = TypedDict(
    "Axis",
    {
        # axis on/off.
        "show": bool,
        # scale key.
        "scale": str,
        # side of chart - 0: top, 1: rgt, 2: btm, 3: lft.
        "side": Literal[0, 1, 2, 3],
        # height of x axis or width of y axis in CSS pixels alloted for values, gap & ticks, but excluding axis label.
        "size": Union[float, JSCallback],
        # gap between axis values and axis baseline (or ticks, if enabled) in CSS pixels.
        "gap": float,
        # font used for axis values.
        "font": CSSValue,
        # font-size multiplier for multi-line axis values (similar to CSS line-height: 1.5em).
        "lineGap": float,
        # color of axis label & values.
        "stroke": CSSValue,
        # axis label text.
        "label": Union[str, JSCallback],
        # height of x axis label or width of y axis label in CSS pixels alloted for label text + labelGap.
        "labelSize": float,
        # gap between label baseline and tick values in CSS pixels.
        "labelGap": float,
        # font used for axis label.
        "labelFont": CSSValue,
        # minimum grid & tick spacing in CSS pixels.
        "space": Union[float, JSCallback],
        # available divisors for axis ticks, values, grid.
        "incrs": Union[Tuple[float, ...], JSCallback],
        # determines how and where the axis must be split for placing ticks, values, grid.
        "splits": Union[Tuple[float, ...], JSCallback],
        # can filter which splits are passed to axis.values() for rendering. e.g splits.map(v => v % 2 == 0 ? v : null).
        "filter": JSCallback,
        # formats values for rendering.
        "values": Union[
            Tuple[Union[str, float, None], ...],
            JSCallback,
            str,
            Tuple[Tuple[Union[str, float, None], ...], ...],
        ],
        # values rotation in degrees off horizontal (only bottom axes w/ side: 2).
        "rotate": Union[float, JSCallback],
        # text alignment of axis values - 1: left, 2: right.
        "align": Literal[1, 2],
        # baseline for text alignment of axis values - 1: inside, 2: outside.
        "alignTo": Literal[1, 2],
        # gridlines to draw from this axis' splits.
        "grid": Axis_Grid,
        # ticks to draw from this axis' splits.
        "ticks": Axis_Ticks,
        # axis border/edge rendering.
        "border": Axis_Border,
    },
    total=False,
)
""":py:class:`~typing.TypedDict` type for uPlot."""

Cursor = TypedDict(
    "Cursor",
    {
        # cursor on/off.
        "show": bool,
        # vertical crosshair on/off.
        "x": bool,
        # horizontal crosshair on/off.
        "y": bool,
        # cursor position left offset in CSS pixels (relative to plotting area).
        "left": float,
        # cursor position top offset in CSS pixels (relative to plotting area).
        "top": float,
        # closest data index to cursor (closestIdx).
        "idx": Union[float, None],
        # returns data idx used for hover points & legend display (defaults to closestIdx).
        "dataIdx": JSCallback,
        # a series-matched array of indices returned by dataIdx().
        "idxs": Tuple[Union[float, None], ...],
        # fires on debounced mousemove events; returns refined [left, top] tuple to snap cursor position.
        "move": JSCallback,
        # series hover points.
        "points": Cursor_Points,
        # event listener proxies (can be overridden to tweak interaction behavior).
        "bind": Cursor_Bind,
        # determines vt/hz cursor dragging to set selection & setScale (zoom).
        "drag": Cursor_Drag,
        # sync cursor between multiple charts.
        "sync": Cursor_Sync,
        # focus series closest to cursor (y).
        "focus": Cursor_Focus,
        # hover data points closest to cursor (x).
        "hover": Cursor_Hover,
        # lock cursor on mouse click in plotting area.
        "lock": bool,
        # the most recent mouse event.
        "event": DOMElement,
    },
    total=False,
)
""":py:class:`~typing.TypedDict` type for uPlot."""

Options = TypedDict(
    "Options",
    {
        "series": Required[Tuple[Series, ...]],
        # 1: aligned & ordered, single-x / y-per-series, 2: unordered & faceted, per-series/per-point x,y,size,label,color,shape,etc.
        "mode": Literal[1, 2],
        # chart title.
        "title": str,
        # id to set on chart div.
        "id": str,
        # className to add to chart div.
        "class": str,
        # initial devicePixelRatio, if different than window.devicePixelRatio.
        "pxRatio": float,
        # data for chart, if none is provided as argument to constructor.
        "data": Union[Tuple[UnknownType, ...], Tuple[()]],
        # converts a unix timestamp to Date that's time-adjusted for the desired timezone.
        "tzDate": JSCallback,
        # creates an efficient formatter for Date objects from a template string, e.g. {YYYY}-{MM}-{DD}.
        "fmtDate": JSCallback,
        # timestamp multiplier that yields 1 millisecond.
        "ms": Literal[1],
        # drawing order for axes/grid & series (default: ["axes", "series"]).
        "drawOrder": Tuple[Literal["axes", "series"], ...],
        # whether vt & hz lines of series/grid/ticks should be crisp/sharp or sub-px antialiased.
        "pxAlign": Union[bool, float],
        "bands": Tuple[Band, ...],
        "scales": Tuple[Union[str, None], Union[str, None]],
        "axes": Tuple[Axis, ...],
        # padding per side, in CSS pixels (can prevent cross-axis labels at the plotting area limits from being chopped off).
        "padding": Tuple[
            Union[float, None, JSCallback],
            Union[float, None, JSCallback],
            Union[float, None, JSCallback],
            Union[float, None, JSCallback],
        ],
        "select": Select,
        "legend": Legend,
        "cursor": Cursor,
        "focus": Focus,
        "hooks": UnknownType,
        "plugins": Tuple[Plugin, ...],
    },
    total=False,
)
""":py:class:`~typing.TypedDict` type for uPlot."""



from __future__ import annotations

import dataclasses
import json
import sys
from pathlib import Path
from typing import Tuple, cast

import imageio.v3 as iio
import numpy as np
import numpy.typing as npt

from ..transforms import SO3


def _resize_nearest(array: npt.NDArray, target_shape: Tuple[int, int]) -> npt.NDArray:
    """Resize array to target shape using nearest neighbor interpolation.

    Args:
        array: Input array with shape (H, W) or (H, W, C).
        target_shape: Target (height, width).

    Returns:
        Resized array with shape (target_shape[0], target_shape[1]) or
        (target_shape[0], target_shape[1], C) if input has channels.
    """
    h_src, w_src = array.shape[:2]
    h_tgt, w_tgt = target_shape

    # Compute indices for nearest neighbor sampling.
    row_indices = (np.arange(h_tgt) * h_src / h_tgt).astype(np.int32)
    col_indices = (np.arange(w_tgt) * w_src / w_tgt).astype(np.int32)

    # Apply indexing to resize.
    return array[row_indices[:, None], col_indices[None, :]]


class Record3dLoader:
    """Helper for loading frames for Record3D captures."""

    # NOTE(hangg): Consider moving this module into
    # `examples/7_record3d_visualizer.py` since it is usecase-specific.

    def __init__(self, data_dir: Path):
        metadata_path = data_dir / "metadata"

        # Read metadata.
        metadata = json.loads(metadata_path.read_text())

        K: np.ndarray = np.array(metadata["K"], np.float32).reshape(3, 3).T
        fps = metadata["fps"]

        T_world_cameras: np.ndarray = np.array(metadata["poses"], np.float32)
        # Convert quaternions (xyzw format) to rotation matrices.
        rotation_matrices = SO3.from_quaternion_xyzw(T_world_cameras[:, :4]).as_matrix()
        T_world_cameras = np.concatenate(
            [
                rotation_matrices,
                T_world_cameras[:, 4:, None],
            ],
            -1,
        )
        T_world_cameras = (T_world_cameras @ np.diag([1, -1, -1, 1])).astype(np.float32)

        self.K = K
        self.fps = fps
        self.T_world_cameras = T_world_cameras

        rgbd_dir = data_dir / "rgbd"
        self.rgb_paths = sorted(rgbd_dir.glob("*.jpg"), key=lambda p: int(p.stem))
        self.depth_paths = [
            rgb_path.with_suffix(".depth") for rgb_path in self.rgb_paths
        ]
        self.conf_paths = [rgb_path.with_suffix(".conf") for rgb_path in self.rgb_paths]

    def num_frames(self) -> int:
        return len(self.rgb_paths)

    def get_frame(self, index: int) -> Record3dFrame:
        # `pyliblzfse` can be hard to install on some Windows systems and is
        # only needed for Record3D, so we don't do a global import.
        try:
            import liblzfse
        except ImportError:
            print("liblzfse is missing. Please install with `pip install pyliblzfse`.")
            sys.exit(1)

        # Read conf.
        conf: np.ndarray = np.frombuffer(
            liblzfse.decompress(self.conf_paths[index].read_bytes()), dtype=np.uint8
        )
        if conf.shape[0] == 640 * 480:
            conf = conf.reshape((640, 480))  # For a FaceID camera 3D Video
        elif conf.shape[0] == 256 * 192:
            conf = conf.reshape((256, 192))  # For a LiDAR 3D Video
        else:
            assert False, f"Unexpected conf shape {conf.shape}"

        # Read depth.
        depth: np.ndarray = np.frombuffer(
            liblzfse.decompress(self.depth_paths[index].read_bytes()), dtype=np.float32
        ).copy()
        if depth.shape[0] == 640 * 480:
            depth = depth.reshape((640, 480))  # For a FaceID camera 3D Video
        elif depth.shape[0] == 256 * 192:
            depth = depth.reshape((256, 192))  # For a LiDAR 3D Video
        else:
            assert False, f"Unexpected depth shape {depth.shape}"

        # Read RGB.
        rgb = iio.imread(self.rgb_paths[index])
        return Record3dFrame(
            K=self.K,
            rgb=rgb,
            depth=depth,
            mask=conf == 2,
            T_world_camera=self.T_world_cameras[index],
        )


@dataclasses.dataclass
class Record3dFrame:
    """A single frame from a Record3D capture."""

    K: npt.NDArray[np.float32]
    rgb: npt.NDArray[np.uint8]
    depth: npt.NDArray[np.float32]
    mask: npt.NDArray[np.bool_]
    T_world_camera: npt.NDArray[np.float32]

    def get_point_cloud(
        self, downsample_factor: int = 1
    ) -> Tuple[npt.NDArray[np.float32], npt.NDArray[np.uint8]]:
        rgb = self.rgb[::downsample_factor, ::downsample_factor]
        target_shape = (rgb.shape[0], rgb.shape[1])
        depth = _resize_nearest(self.depth, target_shape)
        mask = cast(
            npt.NDArray[np.bool_],
            _resize_nearest(self.mask, target_shape),
        )
        assert depth.shape == rgb.shape[:2]

        K = self.K
        T_world_camera = self.T_world_camera

        img_wh = rgb.shape[:2][::-1]

        grid = (
            np.stack(np.meshgrid(np.arange(img_wh[0]), np.arange(img_wh[1])), 2) + 0.5
        )
        grid = grid * downsample_factor

        homo_grid = np.pad(grid[mask], np.array([[0, 0], [0, 1]]), constant_values=1)
        local_dirs = np.einsum("ij,bj->bi", np.linalg.inv(K), homo_grid)
        dirs = np.einsum("ij,bj->bi", T_world_camera[:3, :3], local_dirs)
        points = (T_world_camera[:, -1] + dirs * depth[mask, None]).astype(np.float32)  # type: ignore
        point_colors = rgb[mask]

        return points, point_colors



from __future__ import annotations

import asyncio
import dataclasses
import io
import mimetypes
import os
import threading
import time
import warnings
from collections.abc import Coroutine
from concurrent.futures import ThreadPoolExecutor
from pathlib import Path
from typing import TYPE_CHECKING, Any, Callable, ContextManager, Optional, TypeVar, cast, overload

import imageio.v3 as iio
import numpy as np
import numpy.typing as npt
from typing_extensions import Literal, deprecated

from . import _client_autobuild, _messages, infra
from . import transforms as tf
from ._backwards_compat_shims import DeprecatedAttributeShim
from ._gui_api import GuiApi, LiteralColor, _make_uuid
from ._notification_handle import NotificationHandle, _NotificationHandleState
from ._scene_api import SceneApi, cast_vector
from ._threadpool_exceptions import print_threadpool_errors
from ._tunnel import ViserTunnel
from .infra._infra import StateSerializer


class InitialCameraConfig:
    """Configuration for the initial camera pose.

    Accessed via :attr:`ViserServer.initial_camera`. Values set here determine:

    1. The starting camera pose for new client connections
    2. The pose that "Reset View" returns to in the client

    Default behavior (when properties are not explicitly set):
        The client uses a built-in default camera position that provides a
        reasonable view regardless of the scene's up direction. This default
        is specified in three.js coordinates and does not require world
        coordinate transformation.

    When properties are explicitly set, they are interpreted as viser world
    coordinates and transformed appropriately based on the scene's up direction.

    When properties are changed after clients are connected, only the "Reset
    View" target is updated. Clients' current camera positions are not moved,
    allowing users to continue working undisturbed.

    Note that URL parameters (e.g., ``?initialCameraPosition=1,2,3``) take
    priority over server-set values.

    The API is designed to match :class:`CameraHandle`, which is used for
    per-client camera control.
    """

    def __init__(self, broadcast: Callable[[_messages.Message], None]) -> None:
        self._broadcast = broadcast
        self._position: npt.NDArray[np.float64] = np.array([3.0, 3.0, 3.0])
        self._look_at: npt.NDArray[np.float64] = np.array([0.0, 0.0, 0.0])
        # None means "same as the scene up direction".
        self._up: npt.NDArray[np.float64] | None = None
        # 75 degrees in radians; matches three.js PerspectiveCamera default.
        self._fov: float = 75.0 * np.pi / 180.0
        self._near: float = 0.01
        self._far: float = 1000.0

    @property
    def position(self) -> npt.NDArray[np.float64]:
        """Camera position in world coordinates."""
        return self._position

    @position.setter
    def position(
        self, value: tuple[float, float, float] | npt.NDArray[np.floating]
    ) -> None:
        self._position = np.asarray(value, dtype=np.float64)
        self._broadcast(
            _messages.SetCameraPositionMessage(cast_vector(value, 3), initial=True)
        )

    @property
    def look_at(self) -> npt.NDArray[np.float64]:
        """Point the camera looks at in world coordinates."""
        return self._look_at

    @look_at.setter
    def look_at(
        self, value: tuple[float, float, float] | npt.NDArray[np.floating]
    ) -> None:
        self._look_at = np.asarray(value, dtype=np.float64)
        self._broadcast(
            _messages.SetCameraLookAtMessage(cast_vector(value, 3), initial=True)
        )

    @property
    def up(self) -> npt.NDArray[np.float64] | None:
        """Camera up direction, or None for scene up direction."""
        return self._up

    @up.setter
    def up(self, value: tuple[float, float, float] | npt.NDArray[np.floating]) -> None:
        self._up = np.asarray(value, dtype=np.float64)
        self._broadcast(
            _messages.SetCameraUpDirectionMessage(cast_vector(value, 3), initial=True)
        )

    @property
    def fov(self) -> float:
        """Vertical field of view in radians."""
        return self._fov

    @fov.setter
    def fov(self, value: float) -> None:
        self._fov = float(value)
        self._broadcast(_messages.SetCameraFovMessage(self._fov, initial=True))

    @property
    def near(self) -> float:
        """Near clipping plane distance."""
        return self._near

    @near.setter
    def near(self, value: float) -> None:
        self._near = float(value)
        self._broadcast(_messages.SetCameraNearMessage(self._near, initial=True))

    @property
    def far(self) -> float:
        """Far clipping plane distance."""
        return self._far

    @far.setter
    def far(self, value: float) -> None:
        self._far = float(value)
        self._broadcast(_messages.SetCameraFarMessage(self._far, initial=True))


@dataclasses.dataclass
class _CameraHandleState:
    """Information about a client's camera state."""

    client: ClientHandle
    wxyz: npt.NDArray[np.float64]
    position: npt.NDArray[np.float64]
    fov: float
    image_height: int
    image_width: int
    near: float
    far: float
    look_at: npt.NDArray[np.float64]
    up_direction: npt.NDArray[np.float64]
    update_timestamp: float
    camera_cb: list[Callable[[CameraHandle], None | Coroutine]]


class CameraHandle:
    """A handle for reading and writing the camera state of a particular
    client. Typically accessed via :attr:`ClientHandle.camera`."""

    def __init__(self, client: ClientHandle) -> None:
        self._state = _CameraHandleState(
            client,
            wxyz=np.zeros(4),
            position=np.zeros(3),
            fov=0.0,
            image_height=0,
            image_width=0,
            near=0.01,
            far=1000.0,
            look_at=np.zeros(3),
            up_direction=np.zeros(3),
            update_timestamp=0.0,
            camera_cb=[],
        )

    @property
    def client(self) -> ClientHandle:
        """Client that this camera corresponds to."""
        return self._state.client

    @property
    def wxyz(self) -> npt.NDArray[np.float64]:
        """Corresponds to the R in `P_world = [R | t] p_camera`. Synchronized
        automatically when assigned."""
        assert self._state.update_timestamp != 0.0
        return self._state.wxyz

    # Note: asymmetric properties are supported in Pyright, but not yet in mypy.
    # - https://github.com/python/mypy/issues/3004
    # - https://github.com/python/mypy/pull/11643
    @wxyz.setter
    def wxyz(self, wxyz: tuple[float, float, float, float] | np.ndarray) -> None:
        R_world_camera = tf.SO3(np.asarray(wxyz)).as_matrix()
        look_distance = np.linalg.norm(self.look_at - self.position)

        # We're following OpenCV conventions: look_direction is +Z, up_direction is -Y,
        # right_direction is +X.
        look_direction = R_world_camera[:, 2]
        up_direction = -R_world_camera[:, 1]
        right_direction = R_world_camera[:, 0]

        # Minimize our impact on the orbit controls by keeping the new up direction as
        # close to the old one as possible.
        projected_up_direction = (
            self.up_direction
            - float(self.up_direction @ right_direction) * right_direction
        )
        up_cosine = float(up_direction @ projected_up_direction)
        if abs(up_cosine) < 0.05:
            projected_up_direction = up_direction
        elif up_cosine < 0.0:
            projected_up_direction = up_direction

        new_look_at = look_direction * look_distance + self.position

        # Update lookat and up direction.
        self.look_at = new_look_at
        self.up_direction = projected_up_direction

        # The internal camera orientation should be set in the look_at /
        # up_direction setters. We can uncomment this assert to check this.
        # assert np.allclose(self._state.wxyz, wxyz) or np.allclose(
        #     self._state.wxyz, -wxyz
        # )

    @property
    def position(self) -> npt.NDArray[np.float64]:
        """Corresponds to the t in `P_world = [R | t] p_camera`. Synchronized
        automatically when assigned.

        To preserve the camera orientation, position updates translate both the camera
        and its `look_at` point together. To change position while looking at a fixed
        point, set `look_at` after updating `position`.
        """
        assert self._state.update_timestamp != 0.0
        return self._state.position

    @position.setter
    def position(self, position: tuple[float, float, float] | np.ndarray) -> None:
        position_array = np.asarray(position).astype(np.float64)
        if np.allclose(position_array, self._state.position):
            return
        offset = position_array - np.array(self.position)  # type: ignore
        self._state.position = position_array

        position_tuple = cast_vector(position, 3)
        self._state.client._websock_connection.queue_message(
            _messages.SetCameraPositionMessage(position_tuple)
        )
        self.look_at = np.array(self.look_at) + offset
        self._state.update_timestamp = time.time()

    def _update_wxyz(self) -> None:
        """Compute and update the camera orientation from the internal look_at, position, and up vectors."""
        z = self._state.look_at - self._state.position
        z /= np.linalg.norm(z)
        y = tf.SO3.exp(z * np.pi) @ self._state.up_direction
        y = y - np.dot(z, y) * z
        y /= np.linalg.norm(y)
        x = np.cross(y, z)
        self._state.wxyz = tf.SO3.from_matrix(np.stack([x, y, z], axis=1)).wxyz.astype(
            np.float64
        )

    @property
    def fov(self) -> float:
        """Vertical field of view of the camera, in radians. Synchronized automatically
        when assigned."""
        assert self._state.update_timestamp != 0.0
        return self._state.fov

    @fov.setter
    def fov(self, fov: float) -> None:
        if np.allclose(self._state.fov, fov):
            return
        self._state.fov = fov
        self._state.update_timestamp = time.time()
        self._state.client._websock_connection.queue_message(
            _messages.SetCameraFovMessage(fov)
        )

    @property
    def near(self) -> float:
        """Near clipping plane distance. Synchronized automatically when
        assigned."""
        assert self._state.update_timestamp != 0.0
        return self._state.near

    @near.setter
    def near(self, near: float) -> None:
        if np.allclose(self._state.near, near):
            return
        self._state.near = near
        self._state.update_timestamp = time.time()
        self._state.client._websock_connection.queue_message(
            _messages.SetCameraNearMessage(near)
        )

    @property
    def far(self) -> float:
        """Far clipping plane distance. Synchronized automatically when
        assigned."""
        assert self._state.update_timestamp != 0.0
        return self._state.far

    @far.setter
    def far(self, far: float) -> None:
        if np.allclose(self._state.far, far):
            return
        self._state.far = far
        self._state.update_timestamp = time.time()
        self._state.client._websock_connection.queue_message(
            _messages.SetCameraFarMessage(far)
        )

    @property
    def aspect(self) -> float:
        """Canvas width divided by height. Not assignable."""
        assert self._state.update_timestamp != 0.0
        return float(self._state.image_width) / self._state.image_height

    @property
    def image_height(self) -> int:
        """Image height in pixels. Not assignable."""
        assert self._state.update_timestamp != 0.0
        return self._state.image_height

    @property
    def image_width(self) -> int:
        """Image width in pixels. Not assignable."""
        assert self._state.update_timestamp != 0.0
        return self._state.image_width

    @property
    def update_timestamp(self) -> float:
        assert self._state.update_timestamp != 0.0
        return self._state.update_timestamp

    @property
    def look_at(self) -> npt.NDArray[np.float64]:
        """Look at point for the camera. Synchronized automatically when set."""
        assert self._state.update_timestamp != 0.0
        return self._state.look_at

    @look_at.setter
    def look_at(self, look_at: tuple[float, float, float] | np.ndarray) -> None:
        look_at_array = np.asarray(look_at).astype(np.float64)
        if np.allclose(self._state.look_at, look_at_array):
            return
        self._state.look_at = look_at_array
        self._state.update_timestamp = time.time()
        self._update_wxyz()
        self._state.client._websock_connection.queue_message(
            _messages.SetCameraLookAtMessage(cast_vector(look_at, 3))
        )

    @property
    def up_direction(self) -> npt.NDArray[np.float64]:
        """Up direction for the camera. Synchronized automatically when set."""
        assert self._state.update_timestamp != 0.0
        return self._state.up_direction

    @up_direction.setter
    def up_direction(
        self, up_direction: tuple[float, float, float] | np.ndarray
    ) -> None:
        up_direction_array = np.asarray(up_direction)
        if np.allclose(self._state.up_direction, up_direction_array):
            return
        self._state.up_direction = np.asarray(up_direction_array)
        self._update_wxyz()
        self._state.update_timestamp = time.time()
        self._state.client._websock_connection.queue_message(
            _messages.SetCameraUpDirectionMessage(cast_vector(up_direction, 3))
        )

    def on_update(
        self, callback: Callable[[CameraHandle], NoneOrCoroutine]
    ) -> Callable[[CameraHandle], NoneOrCoroutine]:
        """Attach a callback to run when a new camera message is received.

        The callback can be either a standard function or an async function:
        - Standard functions (def) will be executed in a threadpool.
        - Async functions (async def) will be executed in the event loop.

        Using async functions can be useful for reducing race conditions.
        """
        self._state.camera_cb.append(callback)
        return callback

    def get_render(
        self,
        height: int,
        width: int,
        transport_format: Literal["png", "jpeg"] = "jpeg",
    ) -> np.ndarray:
        """Request a render from a client, block until it's done and received, then
        return it as a numpy array. This is an alias for :meth:`ClientHandle.get_render()`.

        Args:
            height: Height of rendered image. Should be <= the browser height.
            width: Width of rendered image. Should be <= the browser width.
            transport_format: Image transport format. JPEG will return a lossy (H, W, 3) RGB array. PNG will
                return a lossless (H, W, 4) RGBA array, but can cause memory issues on the frontend if called
                too quickly for higher-resolution images.
        """
        return self._state.client.get_render(
            height, width, transport_format=transport_format
        )


NoneOrCoroutine = TypeVar("NoneOrCoroutine", None, Coroutine)


# Don't inherit from RenamedAttributeCompatShim during type checking, because
# this will unnecessarily suppress type errors. (from the overriding of
# __getattr__).
class ClientHandle(DeprecatedAttributeShim if not TYPE_CHECKING else object):
    """A handle is created for each client that connects to a server. Handles can be
    used to communicate with just one client, as well as for reading and writing of
    camera state.

    Similar to :class:`ViserServer`, client handles also expose scene and GUI
    interfaces at :attr:`ClientHandle.scene` and :attr:`ClientHandle.gui`. If
    these are used, for example via a client's
    :meth:`SceneApi.add_point_cloud()` method, created elements are local to
    only one specific client.
    """

    def __init__(
        self, conn: infra.WebsockClientConnection, server: ViserServer
    ) -> None:
        # Private attributes.
        self._websock_connection = conn
        self._viser_server = server

        # Public attributes.
        self.scene: SceneApi = SceneApi(
            self, thread_executor=server._thread_executor, event_loop=server._event_loop
        )
        """Handle for interacting with the 3D scene."""
        self.gui: GuiApi = GuiApi(
            self, thread_executor=server._thread_executor, event_loop=server._event_loop
        )
        """Handle for interacting with the GUI."""
        self.client_id: int = conn.client_id
        """Unique ID for this client."""
        self.camera: CameraHandle = CameraHandle(self)
        """Handle for reading from and manipulating the client's viewport camera."""
        self.browser_info: Optional[_messages.BrowserInfoMessage] = None
        """Browser info for this client."""

    def flush(self) -> None:
        """Flush the outgoing message buffer. Any buffered messages will immediately be
        sent. (by default they are windowed)"""
        self._viser_server._websock_server.flush_client(self.client_id)

    def atomic(self) -> ContextManager[None]:
        """Returns a context where: all outgoing messages are grouped and applied by
        clients atomically.

        This should be treated as a soft constraint that's helpful for things
        like animations, or when we want position and orientation updates to
        happen synchronously.

        Returns:
            Context manager.
        """
        return self._websock_connection.atomic()

    def send_file_download(
        self,
        filename: str,
        content: bytes,
        chunk_size: int = 1024 * 1024,
        save_immediately: bool = False,
    ) -> None:
        """Send a file for a client or clients to download.

        Args:
            filename: Name of the file to send. Used to infer MIME type.
            content: Content of the file.
            chunk_size: Number of bytes to send at a time.
            save_immediately: Whether to save the file immediately. If `False`,
                a link to the file will be shown as a notification. Being able to
                right click the link and choose "Save as..." can be useful.
        """
        mime_type = mimetypes.guess_type(filename, strict=False)[0]
        if mime_type is None:
            mime_type = "application/octet-stream"

        parts = [
            content[i * chunk_size : (i + 1) * chunk_size]
            for i in range(int(np.ceil(len(content) / chunk_size)))
        ]

        uuid = _make_uuid()
        self._websock_connection.queue_message(
            _messages.FileTransferStartDownload(
                save_immediately=save_immediately,
                transfer_uuid=uuid,
                filename=filename,
                mime_type=mime_type,
                part_count=len(parts),
                size_bytes=len(content),
            )
        )
        for i, part in enumerate(parts):
            self._websock_connection.queue_message(
                _messages.FileTransferPart(
                    None,
                    transfer_uuid=uuid,
                    part_index=i,
                    content=part,
                )
            )
            self.flush()

    @overload
    def add_notification(
        self,
        title: str,
        body: str,
        *,
        loading: bool = False,
        with_close_button: bool = True,
        auto_close_seconds: float | None = None,
        color: LiteralColor | tuple[int, int, int] | None = None,
    ) -> NotificationHandle: ...

    @overload
    @deprecated(
        "The `auto_close` argument has been deprecated. Use `auto_close_seconds` instead."
    )
    def add_notification(
        self,
        title: str,
        body: str,
        *,
        loading: bool = False,
        with_close_button: bool = True,
        auto_close: int | Literal[False] = False,
        color: LiteralColor | tuple[int, int, int] | None = None,
    ) -> NotificationHandle: ...

    def add_notification(
        self,
        title: str,
        body: str,
        *,
        loading: bool = False,
        with_close_button: bool = True,
        # In seconds: current API.
        auto_close_seconds: float | None = None,
        # In milliseconds: deprecated.
        auto_close: int | Literal[False] = False,
        color: LiteralColor | tuple[int, int, int] | None = None,
    ) -> NotificationHandle:
        """Add a notification to the client's interface.

        This method creates a new notification that will be displayed at the
        top left corner of the client's viewer. Notifications are useful for
        providing alerts or status updates to users.

        .. deprecated:: 1.0.0
            The `auto_close` argument is deprecated. Use `auto_close_seconds` instead.

        Args:
            title: Title to display on the notification.
            body: Message to display on the notification body.
            loading: Whether the notification shows loading icon.
            with_close_button: Whether the notification can be manually closed.
            auto_close_seconds: Time before the notification automatically
                closes; None if the notification does not close on its own.

        Returns:
            A handle that can be used to interact with the GUI element.
        """
        if auto_close is not False:
            warnings.warn(
                "The `auto_close` (milliseconds) argument has been deprecated. Use `auto_close_seconds` instead.",
                category=DeprecationWarning,
                stacklevel=2,
            )
            auto_close_seconds = auto_close / 1000.0
        handle = NotificationHandle(
            _NotificationHandleState(
                websock_interface=self._websock_connection,
                uuid=_make_uuid(),
                props=_messages.NotificationProps(
                    title=title,
                    body=body,
                    loading=loading,
                    with_close_button=with_close_button,
                    auto_close_seconds=auto_close_seconds,
                    color=color,
                ),
            )
        )
        handle._sync_with_client("show")
        return handle

    @overload
    def get_render(
        self,
        height: int,
        width: int,
        *,
        wxyz: tuple[float, float, float, float] | np.ndarray,
        position: tuple[float, float, float] | np.ndarray,
        fov: float,
        transport_format: Literal["png", "jpeg"] = "jpeg",
    ) -> np.ndarray: ...

    @overload
    def get_render(
        self,
        height: int,
        width: int,
        *,
        transport_format: Literal["png", "jpeg"] = "jpeg",
    ) -> np.ndarray: ...

    def get_render(
        self,
        height: int,
        width: int,
        *,
        wxyz: tuple[float, float, float, float] | np.ndarray | None = None,
        position: tuple[float, float, float] | np.ndarray | None = None,
        fov: float | None = None,
        transport_format: Literal["png", "jpeg"] = "jpeg",
    ) -> np.ndarray:
        """Request a render from a client, block until it's done and received, then
        return it as a numpy array. If wxyz, position, and fov are not provided, the
        current camera state will be used.

        Args:
            height: Height of rendered image. Should be <= the browser height.
            width: Width of rendered image. Should be <= the browser width.
            wxyz: Camera orientation as a quaternion. If not provided, the current camera
                position will be used.
            position: Camera position. If not provided, the current camera position will
                be used.
            fov: Vertical field of view of the camera, in radians. If not provided, the
                current camera position will be used.
            transport_format: Image transport format. JPEG will return a lossy (H, W, 3) RGB array. PNG will
                return a lossless (H, W, 4) RGBA array, but can cause memory issues on the frontend if called
                too quickly for higher-resolution images.
        """

        # Listen for a render reseponse message, which should contain the rendered
        # image.
        render_ready_event = threading.Event()
        out: np.ndarray | None = None

        connection = self._websock_connection

        def got_render_cb(
            client_id: int, message: _messages.GetRenderResponseMessage
        ) -> None:
            del client_id
            connection.unregister_handler(
                _messages.GetRenderResponseMessage, got_render_cb
            )
            nonlocal out
            out = iio.imread(
                io.BytesIO(message.payload),
                extension=f".{transport_format}",
            )
            render_ready_event.set()

        connection.register_handler(_messages.GetRenderResponseMessage, got_render_cb)
        self._websock_connection.queue_message(
            _messages.GetRenderRequestMessage(
                "image/jpeg" if transport_format == "jpeg" else "image/png",
                height=height,
                width=width,
                # Only used for JPEG. The main reason to use a lower quality version
                # value is (unfortunately) to make life easier for the Javascript
                # garbage collector.
                quality=80,
                position=cast_vector(
                    position if position is not None else self.camera.position, 3
                ),
                wxyz=cast_vector(wxyz if wxyz is not None else self.camera.wxyz, 4),
                fov=fov if fov is not None else self.camera.fov,
            )
        )
        render_ready_event.wait()
        assert out is not None
        return out


class ViserServer(DeprecatedAttributeShim if not TYPE_CHECKING else object):
    """:class:`ViserServer` is the main class for working with viser. On
    instantiation, it (a) launches a thread with a web server and (b) provides
    a high-level API for interactive 3D visualization.

    **Core API.** Clients can connect via a web browser, and will be shown two
    components: a 3D scene and a 2D GUI panel. Methods belonging to
    :attr:`ViserServer.scene` can be used to add 3D primitives to the scene.
    Methods belonging to :attr:`ViserServer.gui` can be used to add 2D GUI
    elements.

    **Shared state.** Elements added to the server object, for example via a
    server's :meth:`SceneApi.add_point_cloud` or :meth:`GuiApi.add_button`,
    will have state that's shared and synchronized automatically between all
    connected clients. To show elements that are local to a single client, see
    :attr:`ClientHandle.scene` and :attr:`ClientHandle.gui`.

    Args:
        host: Host to bind server to.
        port: Port to bind server to.
        label: Label shown at the top of the GUI panel.
    """

    # Hide deprecated arguments from docstring and type checkers.
    def __init__(
        self,
        host: str = "0.0.0.0",
        port: int = 8080,
        label: str | None = None,
        verbose: bool = True,
        **_deprecated_kwargs,
    ):
        # Check for port override environment variable
        port_override = os.environ.get("_VISER_PORT_OVERRIDE")
        if port_override is not None:
            try:
                port = int(port_override)
            except ValueError:
                warnings.warn(
                    f"Invalid _VISER_PORT_OVERRIDE value: {port_override}. Using default port {port}."
                )

        # Create server.
        server = infra.WebsockServer(
            host=host,
            port=port,
            message_class=_messages.Message,
            http_server_root=Path(__file__).absolute().parent / "client" / "build",
            verbose=verbose,
            client_api_version=1,
        )
        self._websock_server = server

        _client_autobuild.ensure_client_is_built()

        self._initial_camera = InitialCameraConfig(broadcast=server.queue_message)
        self._connection = server
        self._connected_clients: dict[int, ClientHandle] = {}
        self._client_lock = threading.Lock()
        self._client_connect_cb: list[Callable[[ClientHandle], None | Coroutine]] = []
        self._client_disconnect_cb: list[
            Callable[[ClientHandle], None | Coroutine]
        ] = []

        self._thread_executor = ThreadPoolExecutor(max_workers=32)

        self._client_browser_info_cbs: list[Callable[[ClientHandle, _messages.BrowserInfoMessage], None]] = []

        # Run "garbage collector" on message buffer when new clients connect.
        @server.on_client_connect
        async def _(_: infra.WebsockClientConnection) -> None:
            self._run_garbage_collector()

        # For new clients, register and add a handler for camera messages.
        @server.on_client_connect
        async def _(conn: infra.WebsockClientConnection) -> None:
            client = ClientHandle(conn, server=self)
            first = True

            async def handle_camera_message(
                client_id: infra.ClientId, message: _messages.ViewerCameraMessage
            ) -> None:
                nonlocal first

                assert client_id == client.client_id

                # Update the client's camera.
                client.camera._state = _CameraHandleState(
                    client,
                    np.array(message.wxyz),
                    np.array(message.position),
                    fov=message.fov,
                    image_height=message.image_height,
                    image_width=message.image_width,
                    near=message.near,
                    far=message.far,
                    look_at=np.array(message.look_at),
                    up_direction=np.array(message.up_direction),
                    update_timestamp=time.time(),
                    camera_cb=client.camera._state.camera_cb,
                )

                # We consider a client to be connected after the first camera message is
                # received.
                if first:
                    first = False
                    with self._client_lock:
                        self._connected_clients[conn.client_id] = client
                        for cb in self._client_connect_cb:
                            if asyncio.iscoroutinefunction(cb):
                                await cb(client)
                            else:
                                self._thread_executor.submit(
                                    cb, client
                                ).add_done_callback(print_threadpool_errors)

                for camera_cb in client.camera._state.camera_cb:
                    if asyncio.iscoroutinefunction(camera_cb):
                        await camera_cb(client.camera)
                    else:
                        self._thread_executor.submit(
                            camera_cb, client.camera
                        ).add_done_callback(print_threadpool_errors)

            conn.register_handler(_messages.ViewerCameraMessage, handle_camera_message)

            def handle_browser_info(
                client_id: infra.ClientId, message: _messages.BrowserInfoMessage
            ) -> None:
                assert client_id == client.client_id

                client.browser_info = message
                for cb in self._client_browser_info_cbs:
                    cb(client, message)

            conn.register_handler(_messages.BrowserInfoMessage, handle_browser_info)

        # Remove clients when they disconnect.
        @server.on_client_disconnect
        async def _(conn: infra.WebsockClientConnection) -> None:
            with self._client_lock:
                if conn.client_id not in self._connected_clients:
                    return

                handle = self._connected_clients.pop(conn.client_id)
                for cb in self._client_disconnect_cb:
                    if asyncio.iscoroutinefunction(cb):
                        await cb(handle)
                    else:
                        self._thread_executor.submit(cb, handle).add_done_callback(
                            print_threadpool_errors
                        )

        # Start the server.
        server.start()
        self._event_loop = server._broadcast_buffer.event_loop

        self.scene: SceneApi = SceneApi(
            self, thread_executor=self._thread_executor, event_loop=self._event_loop
        )
        """Handle for interacting with the 3D scene."""

        self.gui: GuiApi = GuiApi(
            self, thread_executor=self._thread_executor, event_loop=self._event_loop
        )
        """Handle for interacting with the GUI."""

        server.register_handler(
            _messages.ShareUrlDisconnect,
            lambda client_id, msg: self.disconnect_share_url(),
        )

        def request_share_url_no_return() -> None:  # To suppress type error.
            self.request_share_url()

        server.register_handler(
            _messages.ShareUrlRequest,
            lambda client_id, msg: cast(None, request_share_url_no_return()),
        )

        # Form status print.
        import rich
        from rich import box, style
        from rich.panel import Panel
        from rich.table import Table

        port = server._port  # Port may have changed.
        if host == "0.0.0.0":
            # 0.0.0.0 is not a real IP and people are often confused by it;
            # we'll just print localhost. This is questionable from a security
            # perspective, but probably fine for our use cases.
            http_url = f"http://localhost:{port}"
            ws_url = f"ws://localhost:{port}"
        else:
            http_url = f"http://{host}:{port}"
            ws_url = f"ws://{host}:{port}"
        table = Table(
            title=None,
            show_header=False,
            box=box.MINIMAL,
            title_style=style.Style(bold=True),
        )
        table.add_row("HTTP", http_url)
        table.add_row("Websocket", ws_url)
        rich.print(
            Panel(
                table,
                title=f"[bold]viser[/bold] [dim](listening *:{port})[/dim]"
                if host == "0.0.0.0"
                else "[bold]viser[/bold]",
                expand=False,
            )
        )

        self._share_tunnel: ViserTunnel | None = None

        # Create share tunnel if requested.
        # This is deprecated: we should use get_share_url() instead.
        share = _deprecated_kwargs.get("share", False)
        if share:
            self.request_share_url()

        self.scene.reset()
        self.scene.set_up_direction("+z")
        self.gui.reset()
        self.gui.set_panel_label(label)

    @property
    def initial_camera(self) -> InitialCameraConfig:
        """Configuration for initial camera pose.

        Set these values to control the initial camera position for new
        clients and serialized/embedded scenes. The API is designed to match
        :class:`viser.CameraHandle`, which is used for per-client camera control.

        Example usage::

            server.initial_camera.position = (5.0, 5.0, 3.0)
            server.initial_camera.look_at = (0.0, 0.0, 0.0)
        """
        return self._initial_camera

    def _run_garbage_collector(self, force: bool = False) -> None:
        """Clean up old messages. This is not elegant; a refactor of our
        message persistence logic will significantly reduce complexity."""
        buffer = self._websock_server._broadcast_buffer
        with buffer.buffer_lock:
            # Skip garbage collection if we have messages that are queeud but
            # not yet processed by the window generators.
            #
            # This makes sure that we don't accidentally cull messages before
            # they're sent to existing clients. RemoveSceneNodeMessage, for example,
            # needs to be sent to old clients but not new ones.
            if (
                not force
                and self._websock_server._broadcast_buffer.message_event.is_set()
            ):
                return

            remove_message_ids: list[int] = []

            remove_scene_names: set[str] = set()
            remove_gui_uuids: set[str] = set()

            for id, message in reversed(buffer.message_from_id.items()):
                # Find scene nodes or GUI elements that were removed.
                if isinstance(message, _messages.RemoveSceneNodeMessage):
                    remove_message_ids.append(id)
                    remove_scene_names.add(message.name)
                elif isinstance(message, _messages.GuiRemoveMessage):
                    remove_message_ids.append(id)
                    remove_gui_uuids.add(message.uuid)
                elif isinstance(message, _messages.GuiCloseModalMessage):
                    remove_message_ids.append(id)

                # For removed elements, no need to send any update messages.
                if (
                    isinstance(
                        message,
                        (
                            _messages.SetPositionMessage,
                            _messages.SetOrientationMessage,
                            _messages.SetBonePositionMessage,
                            _messages.SetBoneOrientationMessage,
                            _messages.SetSceneNodeClickableMessage,
                            _messages.SetSceneNodeVisibilityMessage,
                        ),
                    )
                    and message.name in remove_scene_names
                ):
                    remove_message_ids.append(id)

                if (
                    isinstance(message, _messages.GuiUpdateMessage)
                    and message.uuid in remove_gui_uuids
                ):
                    remove_message_ids.append(id)

            # Remove old messages.
            for id in remove_message_ids:
                message = buffer.message_from_id.pop(id)
                buffer.id_from_redundancy_key.pop(message.redundancy_key())

    def get_host(self) -> str:
        """Returns the host address of the Viser server.

        Returns:
            Host address as string.
        """
        return self._websock_server._host

    def get_port(self) -> int:
        """Returns the port of the Viser server. This could be different from the
        originally requested one.

        Returns:
            Port as integer.
        """
        return self._websock_server._port

    def request_share_url(self, verbose: bool = True) -> str | None:
        """Request a share URL for the Viser server, which allows for public access.
        On the first call, will block until a connecting with the share URL server is
        established. Afterwards, the URL will be returned directly.

        This is an experimental feature that relies on an external server; it shouldn't
        be relied on for critical applications.

        Returns:
            Share URL as string, or None if connection fails or is closed.
        """
        if self._share_tunnel is not None:
            # Tunnel already exists.
            while self._share_tunnel.get_status() in ("ready", "connecting"):
                time.sleep(0.05)
            return self._share_tunnel.get_url()
        else:
            # Create a new tunnel!.
            if verbose:
                import rich

                rich.print("[bold](viser)[/bold] Share URL requested!")

            connect_event = threading.Event()

            self._share_tunnel = ViserTunnel(
                "share.viser.studio", self._websock_server._port
            )

            @self._share_tunnel.on_disconnect
            def _() -> None:
                import rich

                rich.print("[bold](viser)[/bold] Disconnected from share URL")
                self._share_tunnel = None
                self._websock_server.queue_message(_messages.ShareUrlUpdated(None))

            @self._share_tunnel.on_connect
            def _(max_clients: int) -> None:
                assert self._share_tunnel is not None
                share_url = self._share_tunnel.get_url()
                if verbose:
                    import rich

                    if share_url is None:
                        rich.print("[bold](viser)[/bold] Could not generate share URL")
                    else:
                        rich.print(
                            f"[bold](viser)[/bold] Generated share URL (expires in 24 hours, max {max_clients} clients): {share_url}"
                        )
                self._websock_server.queue_message(_messages.ShareUrlUpdated(share_url))
                connect_event.set()

            connect_event.wait()

            url = self._share_tunnel.get_url()
            return url

    def disconnect_share_url(self) -> None:
        """Disconnect from the share URL server."""
        if self._share_tunnel is not None:
            self._share_tunnel.close()
        else:
            import rich

            rich.print(
                "[bold](viser)[/bold] Tried to disconnect from share URL, but already disconnected"
            )

    def stop(self) -> None:
        """Stop the Viser server and associated threads and tunnels."""
        self._websock_server.stop()
        if self._share_tunnel is not None:
            self._share_tunnel.close()

    def get_clients(self) -> dict[int, ClientHandle]:
        """Creates and returns a copy of the mapping from connected client IDs to
        handles.

        Returns:
            Dictionary of clients.
        """
        with self._client_lock:
            return self._connected_clients.copy()

    def on_client_connect(
        self, cb: Callable[[ClientHandle], NoneOrCoroutine]
    ) -> Callable[[ClientHandle], NoneOrCoroutine]:
        """Attach a callback to run for newly connected clients.

        The callback can be either a standard function or an async function:
        - Standard functions (def) will be executed in a threadpool.
        - Async functions (async def) will be executed in the event loop.

        Using async functions can be useful for reducing race conditions.
        """
        with self._client_lock:
            clients = self._connected_clients.copy().values()
            self._client_connect_cb.append(cb)

        # Trigger callback on any already-connected clients.
        # If we have:
        #
        #     server = viser.ViserServer()
        #     server.on_client_connect(...)
        #
        # This makes sure that the the callback is applied to any clients that
        # connect between the two lines.
        for client in clients:
            if asyncio.iscoroutinefunction(cb):
                self._event_loop.create_task(cb(client))
            else:
                self._thread_executor.submit(cb, client).add_done_callback(
                    print_threadpool_errors
                )

        return cb  # type: ignore

    def on_client_disconnect(
        self, cb: Callable[[ClientHandle], NoneOrCoroutine]
    ) -> Callable[[ClientHandle], NoneOrCoroutine]:
        """Attach a callback to run when clients disconnect.

        The callback can be either a standard function or an async function:
        - Standard functions (def) will be executed in a threadpool.
        - Async functions (async def) will be executed in the event loop.

        Using async functions can be useful for reducing race conditions.
        """
        self._client_disconnect_cb.append(cb)
        return cb

    def flush(self) -> None:
        """Flush the outgoing message buffer. Any buffered messages will immediately be
        sent. (by default they are windowed)"""
        self._websock_server.flush()

    def atomic(self) -> ContextManager[None]:
        """Returns a context where: all outgoing messages are grouped and applied by
        clients atomically.

        This should be treated as a soft constraint that's helpful for things
        like animations, or when we want position and orientation updates to
        happen synchronously.

        Returns:
            Context manager.
        """
        return self._websock_server.atomic()

    def send_file_download(
        self,
        filename: str,
        content: bytes,
        chunk_size: int = 1024 * 1024,
        save_immediately: bool = False,
    ) -> None:
        """Send a file for a client or clients to download.

        Args:
            filename: Name of the file to send. Used to infer MIME type.
            content: Content of the file.
            chunk_size: Number of bytes to send at a time.
            save_immediately: Whether to save the file immediately. If `False`,
                a link to the file will be shown as a notification. Being able to
                right click the link and choose "Save as..." can be useful.
        """
        for client in self.get_clients().values():
            client.send_file_download(filename, content, chunk_size, save_immediately)

    def get_event_loop(self) -> asyncio.AbstractEventLoop:
        """Get the asyncio event loop used by the Viser background thread. This
        can be useful for safe concurrent operations."""
        return self._event_loop

    def sleep_forever(self) -> None:
        """Equivalent to:

        while True:
            time.sleep(3600)
        """
        while True:
            time.sleep(3600)

    def _start_scene_recording(self) -> Any:
        """**Old API.**"""
        warnings.warn(
            "_start_scene_recording() has been renamed. See notes in https://github.com/viser-project/viser/pull/357 for the new API.",
            stacklevel=2,
        )

        serializer = self.get_scene_serializer()

        # We'll add a shim for the old API for now. We can remove this later.
        class _SceneRecordCompatibilityShim:
            def set_loop_start(self):
                warnings.warn(
                    "_start_scene_recording() has been renamed. See notes in https://github.com/viser-project/viser/pull/357 for the new API.",
                    stacklevel=2,
                )

            def insert_sleep(self, duration: float):
                warnings.warn(
                    "_start_scene_recording() has been renamed. See notes in https://github.com/viser-project/viser/pull/357 for the new API.",
                    stacklevel=2,
                )
                serializer.insert_sleep(duration)

            def end_and_serialize(self) -> bytes:
                warnings.warn(
                    "_start_scene_recording() has been renamed. See notes in https://github.com/viser-project/viser/pull/357 for the new API.",
                    stacklevel=2,
                )
                return serializer.serialize()

        return _SceneRecordCompatibilityShim()

    def get_scene_serializer(self) -> StateSerializer:
        """Get handle for serializing the scene state.

        This can be used for saving .viser files, which are used for offline
        visualization.
        """
        serializer = self._websock_server.get_message_serializer(
            # Don't record GUI messages. This feels brittle.
            filter=lambda message: "Gui" not in type(message).__name__
        )
        # Insert current scene state.
        for message in self._websock_server._broadcast_buffer.message_from_id.values():
            serializer._insert_message(message)
        return serializer

    def on_browser_info(self, cb: Callable[[ClientHandle, _messages.BrowserInfoMessage], None]) -> Callable[[ClientHandle, _messages.BrowserInfoMessage], None]:
        """Register a callback for BrowserInfoMessages from clients.

        Args:
            cb: Callback function.

        Returns:
            Callback function.
        """
        self._client_browser_info_cbs.append(cb)
        return cb


"""Colmap utilities."""

from ._colmap_utils import read_cameras_binary as read_cameras_binary
from ._colmap_utils import read_cameras_text as read_cameras_text
from ._colmap_utils import read_images_binary as read_images_binary
from ._colmap_utils import read_images_text as read_images_text
from ._colmap_utils import read_points3d_binary as read_points3d_binary
from ._colmap_utils import read_points3D_text as read_points3D_text



from __future__ import annotations

import asyncio
import builtins
import colorsys
import dataclasses
import functools
import threading
import time
from asyncio import AbstractEventLoop
from concurrent.futures import ThreadPoolExecutor
from pathlib import Path
from typing import (
    TYPE_CHECKING,
    Any,
    Protocol,
    Sequence,
    Tuple,
    TypeVar,
    cast,
    overload,
)

import numpy as np
from typing_extensions import (
    Literal,
    LiteralString,
    TypeAlias,
    TypedDict,
    get_type_hints,
)

from viser import theme
from viser._backwards_compat_shims import deprecated_positional_shim

from . import _messages, uplot
from ._gui_handles import (
    GuiButtonGroupHandle,
    GuiButtonHandle,
    GuiCheckboxHandle,
    GuiContainerProtocol,
    GuiDropdownHandle,
    GuiEvent,
    GuiFolderHandle,
    GuiHtmlHandle,
    GuiImageHandle,
    GuiMarkdownHandle,
    GuiModalHandle,
    GuiMultiSliderHandle,
    GuiNumberHandle,
    GuiPlotlyHandle,
    GuiProgressBarHandle,
    GuiRgbaHandle,
    GuiRgbHandle,
    GuiSliderHandle,
    GuiTabGroupHandle,
    GuiTextHandle,
    GuiUploadButtonHandle,
    GuiUplotHandle,
    GuiVector2Handle,
    GuiVector3Handle,
    SupportsRemoveProtocol,
    UploadedFile,
    _GuiButtonHandleState,
    _GuiHandleState,
    _GuiInputHandle,
    _make_uuid,
)
from ._icons import svg_from_icon
from ._icons_enum import IconName
from ._messages import FileTransferPartAck, GuiBaseProps, GuiSliderMark
from ._scene_api import cast_vector
from ._threadpool_exceptions import print_threadpool_errors

if TYPE_CHECKING:
    import plotly.graph_objects as go

    from ._viser import ClientHandle, ViserServer
    from .infra import ClientId

GuiInputPropsType = TypeVar("GuiInputPropsType", bound=GuiBaseProps)
IntOrFloat = TypeVar("IntOrFloat", int, float)
TString = TypeVar("TString", bound=str)
TLiteralString = TypeVar("TLiteralString", bound=LiteralString)
T = TypeVar("T")
LengthTenStrTuple: TypeAlias = Tuple[str, str, str, str, str, str, str, str, str, str]
LiteralColor: TypeAlias = Literal[
    "dark",
    "gray",
    "red",
    "pink",
    "grape",
    "violet",
    "indigo",
    "blue",
    "cyan",
    "green",
    "lime",
    "yellow",
    "orange",
    "teal",
]


def _hex_from_hls(h: float, l: float, s: float) -> str:
    """Converts HLS values in [0.0, 1.0] to a hex-formatted string, eg 0xffffff."""
    return "#" + "".join(
        [
            int(min(255, max(0, channel * 255.0)) + 0.5).to_bytes(1, "little").hex()
            for channel in colorsys.hls_to_rgb(h, l, s)
        ]
    )


def _compute_step(x: float | None) -> float:  # type: ignore
    """For number inputs: compute an increment size from some number.

    Example inputs/outputs:
        100 => 1
        12 => 1
        12.1 => 0.1
        12.02 => 0.01
        0.004 => 0.001
    """
    return 1 if x is None else 10 ** (-_compute_precision_digits(x))


def _compute_precision_digits(x: float) -> int:
    """For number inputs: compute digits of precision from some number.

    Example inputs/outputs:
        100 => 0
        12 => 0
        12.1 => 1
        10.2 => 1
        0.007 => 3
    """
    digits = 0
    while x != round(x, ndigits=digits) and digits < 7:
        digits += 1
    return digits


@dataclasses.dataclass
class _RootGuiContainer:
    _children: dict[str, SupportsRemoveProtocol]


_global_order_counter = 0


def _apply_default_order(order: float | None) -> float:
    """Apply default ordering logic for GUI elements.

    If `order` is set to a float, this function is a no-op and returns it back.
    Otherwise, we increment and return the value of a global counter.
    """
    if order is not None:
        return order

    global _global_order_counter
    _global_order_counter += 1
    return _global_order_counter


@functools.lru_cache(maxsize=None)
def get_type_hints_cached(cls: type[Any]) -> dict[str, Any]:
    return get_type_hints(cls)  # type: ignore


class _FileUploadState(TypedDict):
    filename: str
    mime_type: str
    part_count: int
    parts: dict[int, bytes]
    total_bytes: int
    transferred_bytes: int
    lock: threading.Lock


class GuiApi:
    """Interface for working with the 2D GUI in viser.

    Used by both our global server object, for sharing the same GUI elements
    with all clients, and by individual client handles."""

    _target_container_from_thread_id: dict[int, str] = {}
    """ID of container to put GUI elements into."""

    def __init__(
        self,
        owner: ViserServer | ClientHandle,  # Who do I belong to?
        thread_executor: ThreadPoolExecutor,
        event_loop: AbstractEventLoop,
    ) -> None:
        from ._viser import ViserServer

        self._owner = owner
        """Entity that owns this API."""
        self._thread_executor = thread_executor
        self._event_loop = event_loop

        self._websock_interface = (
            owner._websock_server
            if isinstance(owner, ViserServer)
            else owner._websock_connection
        )
        """Interface for sending and listening to messages."""

        self._gui_input_handle_from_uuid: dict[str, _GuiInputHandle[Any]] = {}
        self._container_handle_from_uuid: dict[str, GuiContainerProtocol] = {
            "root": _RootGuiContainer({})
        }
        self._modal_handle_from_uuid: dict[str, GuiModalHandle] = {}
        self._current_file_upload_states: dict[str, _FileUploadState] = {}

        # Set to True when plotly.min.js has been sent to client.
        self._setup_plotly_js: bool = False

        self._websock_interface.register_handler(
            _messages.GuiUpdateMessage, self._handle_gui_updates
        )
        self._websock_interface.register_handler(
            _messages.GuiButtonHoldMessage, self._handle_gui_button_hold
        )
        self._websock_interface.register_handler(
            _messages.FileTransferStartUpload, self._handle_file_transfer_start
        )
        self._websock_interface.register_handler(
            _messages.FileTransferPart,
            self._handle_file_transfer_part,
        )

    async def _handle_gui_updates(
        self, client_id: ClientId, message: _messages.GuiUpdateMessage
    ) -> None:
        """Callback for handling GUI messages."""
        handle = self._gui_input_handle_from_uuid.get(message.uuid, None)
        if handle is None:
            return
        handle_state = handle._impl

        has_changed = False
        updates_cast = {}
        for prop_name, prop_value in message.updates.items():
            assert hasattr(handle_state, prop_name)
            current_value = getattr(handle_state, prop_name)

            # Do some type casting. This is brittle, but necessary (1) when we
            # expect floats but the Javascript side gives us integers or (2)
            # when we expect tuples but the Javascript side gives us lists.
            if prop_name == "value":
                if isinstance(handle_state.value, tuple):
                    if len(handle_state.value) > 0:
                        # We currently assume non-empty tuple types have length
                        # greater than 0, and contents are all the same type.
                        typ = type(handle_state.value[0])
                        assert all([type(x) == typ for x in handle_state.value])
                        prop_value = tuple([typ(new) for new in prop_value])
                    else:
                        # Empty tuple.
                        prop_value = tuple(prop_value)
                else:
                    prop_value = type(handle_state.value)(prop_value)

            # Update handle property.
            if current_value != prop_value:
                has_changed = True
                setattr(handle_state, prop_name, prop_value)

            # Save value, which might have been cast.
            updates_cast[prop_name] = prop_value

        # Only call update when value has actually changed.
        if not handle_state.is_button and not has_changed:
            return

        # GUI element has been updated!
        handle_state.update_timestamp = time.time()
        for cb in handle_state.update_cb:
            from ._viser import ClientHandle, ViserServer

            # Get the handle of the client that triggered this event.
            if isinstance(self._owner, ClientHandle):
                client = self._owner
            elif isinstance(self._owner, ViserServer):
                client = self._owner._connected_clients.get(client_id, None)
                if client is None:
                    return
            else:
                assert False

            if asyncio.iscoroutinefunction(cb):
                await cb(GuiEvent(client, client_id, handle))
            else:
                self._thread_executor.submit(
                    cb, GuiEvent(client, client_id, handle)
                ).add_done_callback(print_threadpool_errors)

        if handle_state.sync_cb is not None:
            handle_state.sync_cb(client_id, updates_cast)

    async def _handle_gui_button_hold(
        self, client_id: ClientId, message: _messages.GuiButtonHoldMessage
    ) -> None:
        """Callback for handling button hold messages."""
        handle = self._gui_input_handle_from_uuid.get(message.uuid, None)
        if handle is None:
            return

        # Ensure this is a button handle with hold callbacks.
        if not isinstance(handle, GuiButtonHandle):
            return

        # Get callbacks registered for this frequency.
        callbacks = handle._button_impl.hold_cbs_from_freq.get(message.frequency, [])
        if not callbacks:
            return

        # Get the client handle.
        from ._viser import ClientHandle, ViserServer

        if isinstance(self._owner, ClientHandle):
            client = self._owner
        elif isinstance(self._owner, ViserServer):
            client = self._owner._connected_clients.get(client_id, None)
            if client is None:
                return
        else:
            assert False

        # Call all callbacks for this frequency.
        for cb in callbacks:
            if asyncio.iscoroutinefunction(cb):
                await cb(GuiEvent(client, client_id, handle))
            else:
                self._thread_executor.submit(
                    cb, GuiEvent(client, client_id, handle)
                ).add_done_callback(print_threadpool_errors)

    def _handle_file_transfer_start(
        self, client_id: ClientId, message: _messages.FileTransferStartUpload
    ) -> None:
        if message.source_component_uuid not in self._gui_input_handle_from_uuid:
            return
        self._current_file_upload_states[message.transfer_uuid] = {
            "filename": message.filename,
            "mime_type": message.mime_type,
            "part_count": message.part_count,
            "parts": {},
            "total_bytes": message.size_bytes,
            "transferred_bytes": 0,
            "lock": threading.Lock(),
        }

    def _handle_file_transfer_part(
        self, client_id: ClientId, message: _messages.FileTransferPart
    ) -> None:
        if message.transfer_uuid not in self._current_file_upload_states:
            return
        assert message.source_component_uuid in self._gui_input_handle_from_uuid

        state = self._current_file_upload_states[message.transfer_uuid]
        state["parts"][message.part_index] = message.content
        total_bytes = state["total_bytes"]

        with state["lock"]:
            state["transferred_bytes"] += len(message.content)

            # Send ack to the server.
            self._websock_interface.queue_message(
                FileTransferPartAck(
                    source_component_uuid=message.source_component_uuid,
                    transfer_uuid=message.transfer_uuid,
                    transferred_bytes=state["transferred_bytes"],
                    total_bytes=total_bytes,
                )
            )

            if state["transferred_bytes"] < total_bytes:
                return

        # Finish the upload.
        assert state["transferred_bytes"] == total_bytes
        state = self._current_file_upload_states.pop(message.transfer_uuid)

        handle = self._gui_input_handle_from_uuid.get(
            message.source_component_uuid, None
        )
        if handle is None:
            return

        handle_state = handle._impl

        value = UploadedFile(
            name=state["filename"],
            content=b"".join(state["parts"][i] for i in range(state["part_count"])),
        )

        # Update state.
        handle_state.value = value
        handle_state.update_timestamp = time.time()

        # Trigger callbacks.
        for cb in handle_state.update_cb:
            from ._viser import ClientHandle, ViserServer

            # Get the handle of the client that triggered this event.
            if isinstance(self._owner, ClientHandle):
                client = self._owner
            elif isinstance(self._owner, ViserServer):
                client = self._owner._connected_clients.get(client_id, None)
                if client is None:
                    return
            else:
                assert False

            if asyncio.iscoroutinefunction(cb):
                self._event_loop.create_task(cb(GuiEvent(client, client_id, handle)))
            else:
                self._thread_executor.submit(
                    cb, GuiEvent(client, client_id, handle)
                ).add_done_callback(print_threadpool_errors)

    def _get_container_uuid(self) -> str:
        """Get container ID associated with the current thread."""
        return self._target_container_from_thread_id.get(threading.get_ident(), "root")

    def _set_container_uuid(self, container_uuid: str) -> None:
        """Set container ID associated with the current thread."""
        self._target_container_from_thread_id[threading.get_ident()] = container_uuid

    def reset(self) -> None:
        """Reset the GUI."""
        root_container = self._container_handle_from_uuid["root"]
        while len(root_container._children) > 0:
            next(iter(root_container._children.values())).remove()
        while len(self._modal_handle_from_uuid) > 0:
            next(iter(self._modal_handle_from_uuid.values())).close()

    def set_panel_label(self, label: str | None) -> None:
        """Set the main label that appears in the GUI panel.

        Args:
            label: The new label.
        """
        self._websock_interface.queue_message(_messages.SetGuiPanelLabelMessage(label))

    def configure_theme(
        self,
        *,
        titlebar_content: theme.TitlebarConfig | None = None,
        control_layout: Literal["floating", "collapsible", "fixed"] = "floating",
        control_width: Literal["small", "medium", "large"] = "medium",
        dark_mode: bool = False,
        show_logo: bool = True,
        show_share_button: bool = True,
        brand_color: tuple[int, int, int] | None = None,
    ) -> None:
        """Configures the visual appearance of the viser front-end.

        Args:
            titlebar_content: Optional configuration for the title bar.
            control_layout: The layout of control elements, options are "floating",
                            "collapsible", or "fixed".
            control_width: The width of control elements, options are "small",
                           "medium", or "large".
            dark_mode: A boolean indicating if dark mode should be enabled.
            show_logo: A boolean indicating if the logo should be displayed.
            show_share_button: A boolean indicating if the share button should be displayed.
            brand_color: An optional tuple of integers (RGB) representing the brand color.
        """

        colors_cast: LengthTenStrTuple | None = None

        if brand_color is not None:
            assert len(brand_color) in (3, 10)
            if len(brand_color) == 3:
                assert all(map(lambda val: isinstance(val, int), brand_color)), (
                    "All channels should be integers."
                )

                # RGB => HLS.
                h, l, s = colorsys.rgb_to_hls(
                    brand_color[0] / 255.0,
                    brand_color[1] / 255.0,
                    brand_color[2] / 255.0,
                )

                # Automatically generate a 10-color palette.
                min_l = max(l - 0.08, 0.0)
                max_l = min(0.8 + 0.5, 0.9)
                l = max(min_l, min(max_l, l))

                primary_index = 8
                ls = tuple(
                    np.interp(
                        x=np.arange(10),
                        xp=np.array([0, primary_index, 9]),
                        fp=np.array([max_l, l, min_l]),
                    )
                )
                colors_cast = cast(
                    LengthTenStrTuple,
                    tuple(_hex_from_hls(h, float(ls[i]), s) for i in range(10)),
                )

        assert colors_cast is None or all(
            [isinstance(val, str) and val.startswith("#") for val in colors_cast]
        ), "All string colors should be in hexadecimal + prefixed with #, eg #ffffff."

        self._websock_interface.queue_message(
            _messages.ThemeConfigurationMessage(
                titlebar_content=titlebar_content,
                control_layout=control_layout,
                control_width=control_width,
                dark_mode=dark_mode,
                show_logo=show_logo,
                show_share_button=show_share_button,
                colors=colors_cast,
            ),
        )

    @deprecated_positional_shim
    def add_folder(
        self,
        label: str,
        *,
        order: float | None = None,
        expand_by_default: bool = True,
        visible: bool = True,
    ) -> GuiFolderHandle:
        """Add a folder, and return a handle that can be used to populate it.

        Args:
            label: Label to display on the folder.
            order: Optional ordering, smallest values will be displayed first.
            expand_by_default: Open the folder by default. Set to False to collapse it by
                default.
            visible: Whether the component is visible.

        Returns:
            A handle that can be used as a context to populate the folder.
        """
        folder_container_id = _make_uuid()
        order = _apply_default_order(order)
        props = _messages.GuiFolderProps(
            order=order,
            label=label,
            expand_by_default=expand_by_default,
            visible=visible,
        )
        self._websock_interface.queue_message(
            _messages.GuiFolderMessage(
                uuid=folder_container_id,
                container_uuid=self._get_container_uuid(),
                props=props,
            )
        )
        return GuiFolderHandle(
            _GuiHandleState(
                folder_container_id,
                self,
                None,
                props=props,
                parent_container_id=self._get_container_uuid(),
            )
        )

    @deprecated_positional_shim
    def add_modal(
        self,
        title: str,
        *,
        order: float | None = None,
    ) -> GuiModalHandle:
        """Show a modal window, which can be useful for popups and messages, then return
        a handle that can be used to populate it.

        Args:
            title: Title to display on the modal.
            order: Optional ordering, smallest values will be displayed first.

        Returns:
            A handle that can be used as a context to populate the modal.
        """
        modal_container_id = _make_uuid()
        order = _apply_default_order(order)
        self._websock_interface.queue_message(
            _messages.GuiModalMessage(
                order=order,
                uuid=modal_container_id,
                title=title,
            )
        )
        return GuiModalHandle(
            _gui_api=self,
            _uuid=modal_container_id,
        )

    @deprecated_positional_shim
    def add_tab_group(
        self,
        *,
        order: float | None = None,
        visible: bool = True,
    ) -> GuiTabGroupHandle:
        """Add a tab group.

        Args:
            order: Optional ordering, smallest values will be displayed first.
            visible: Whether the component is visible.

        Returns:
            A handle that can be used as a context to populate the tab group.
        """
        tab_group_id = _make_uuid()
        order = _apply_default_order(order)

        message = _messages.GuiTabGroupMessage(
            uuid=tab_group_id,
            container_uuid=self._get_container_uuid(),
            props=_messages.GuiTabGroupProps(
                order=order,
                _tab_labels=(),
                visible=visible,
                _tab_icons_html=(),
                _tab_container_ids=(),
            ),
        )
        self._websock_interface.queue_message(message)
        return GuiTabGroupHandle(
            _GuiHandleState(
                message.uuid,
                self,
                value=None,
                props=message.props,
                parent_container_id=message.container_uuid,
            )
        )

    @deprecated_positional_shim
    def add_markdown(
        self,
        content: str,
        *,
        image_root: Path | None = None,
        order: float | None = None,
        visible: bool = True,
    ) -> GuiMarkdownHandle:
        """Add markdown to the GUI.

        Args:
            content: Markdown content to display.
            image_root: Optional root directory to resolve relative image paths.
            order: Optional ordering, smallest values will be displayed first.
            visible: Whether the component is visible.

        Returns:
            A handle that can be used to interact with the GUI element.
        """
        message = _messages.GuiMarkdownMessage(
            uuid=_make_uuid(),
            container_uuid=self._get_container_uuid(),
            props=_messages.GuiMarkdownProps(
                order=_apply_default_order(order),
                _markdown="",
                visible=visible,
            ),
        )
        self._websock_interface.queue_message(message)

        handle = GuiMarkdownHandle(
            _GuiHandleState(
                message.uuid,
                self,
                None,
                props=message.props,
                parent_container_id=message.container_uuid,
            ),
            _content=content,
            _image_root=image_root,
        )

        # Logic for processing markdown, handling images, etc is all in the
        # `.content` setter, which should send a GuiUpdateMessage.
        handle.content = content
        return handle

    @deprecated_positional_shim
    def add_html(
        self,
        content: str,
        *,
        order: float | None = None,
        visible: bool = True,
    ) -> GuiHtmlHandle:
        """Add HTML to the GUI.

        Args:
            content: HTML content to display.
            order: Optional ordering, smallest values will be displayed first.
            visible: Whether the component is visible.

        Returns:
            A handle that can be used to interact with the GUI element.
        """
        message = _messages.GuiHtmlMessage(
            uuid=_make_uuid(),
            container_uuid=self._get_container_uuid(),
            props=_messages.GuiHtmlProps(
                order=_apply_default_order(order),
                content=content,
                visible=visible,
            ),
        )
        self._websock_interface.queue_message(message)

        handle = GuiHtmlHandle(
            _GuiHandleState(
                message.uuid,
                self,
                None,
                props=message.props,
                parent_container_id=message.container_uuid,
            ),
        )
        return handle

    @deprecated_positional_shim
    def add_image(
        self,
        image: np.ndarray,
        *,
        label: str | None = None,
        format: Literal["auto", "png", "jpeg"] = "auto",
        jpeg_quality: int | None = None,
        order: float | None = None,
        visible: bool = True,
    ) -> GuiImageHandle:
        """Add an image element to the GUI.

        Args:
            image: A numpy array representing the image to display.
            label: Label to display on the image element.
            format: Format to transport and display the image using. 'auto' will use PNG for RGBA images and JPEG for RGB.
            jpeg_quality: Quality of the jpeg image (if jpeg format is used).
            order: Order of the element for sorting.
            visible: Whether the image element is visible initially.

        Returns:
            Handle for manipulating the image element.
        """
        # Resolve format if auto.
        if format == "auto":
            resolved_format = "png" if image.shape[2] == 4 else "jpeg"
        else:
            resolved_format = format

        message = _messages.GuiImageMessage(
            uuid=_make_uuid(),
            container_uuid=self._get_container_uuid(),
            props=_messages.GuiImageProps(
                _data=None,  # Sent in prop update later.
                label=label,
                _format=resolved_format,
                order=_apply_default_order(order),
                visible=visible,
            ),
        )
        self._websock_interface.queue_message(message)

        handle = GuiImageHandle(
            _GuiHandleState(
                message.uuid,
                self,
                None,
                props=message.props,
                parent_container_id=message.container_uuid,
            ),
            _image=image,
            _jpeg_quality=jpeg_quality,
        )
        handle._user_format = format
        handle.image = image
        return handle

    @deprecated_positional_shim
    def add_plotly(
        self,
        figure: go.Figure,
        *,
        aspect: float = 1.0,
        order: float | None = None,
        visible: bool = True,
    ) -> GuiPlotlyHandle:
        """Add a Plotly figure to the GUI. Requires the `plotly` package to be
        installed.

        .. note::
           Updates to Plotly figures can be slow when you have many plots or frequent updates. For real-time visualization, consider using
           :meth:`add_uplot` instead.

        Args:
            figure: Plotly figure to display.
            aspect: Aspect ratio of the plot in the control panel (width/height).
            order: Optional ordering, smallest values will be displayed first.
            visible: Whether the component is visible.

        Returns:
            A handle that can be used to interact with the GUI element.
        """

        # If plotly.min.js hasn't been sent to the client yet, the client won't be able
        # to render the plot. Send this large file now! (~3MB)
        if not self._setup_plotly_js:
            # Check if plotly is installed.
            try:
                import plotly
            except ImportError:
                raise ImportError(
                    "You must have the `plotly` package installed to use the Plotly GUI element."
                )

            # Check that plotly.min.js exists.
            plotly_path = (
                Path(plotly.__file__).parent / "package_data" / "plotly.min.js"
            )
            assert plotly_path.exists(), (
                f"Could not find plotly.min.js at {plotly_path}."
            )

            # Send it over!
            plotly_js = plotly_path.read_text(encoding="utf-8")
            self._websock_interface.queue_message(
                _messages.RunJavascriptMessage(source=plotly_js)
            )

            # Update the flag so we don't send it again.
            self._setup_plotly_js = True

        # After plotly.min.js has been sent, we can send the plotly figure.
        # Empty string for `plotly_json_str` is a signal to the client to render nothing.
        message = _messages.GuiPlotlyMessage(
            uuid=_make_uuid(),
            container_uuid=self._get_container_uuid(),
            props=_messages.GuiPlotlyProps(
                order=_apply_default_order(order),
                _plotly_json_str="",
                aspect=1.0,
                visible=visible,
            ),
        )
        self._websock_interface.queue_message(message)

        handle = GuiPlotlyHandle(
            _GuiHandleState(
                message.uuid,
                self,
                value=None,
                props=message.props,
                parent_container_id=message.container_uuid,
            ),
            _figure=figure,
        )

        # Set the plotly handle properties.
        handle.figure = figure
        handle.aspect = aspect
        return handle

    @deprecated_positional_shim
    def add_uplot(
        self,
        data: tuple[np.ndarray, ...],
        series: tuple[uplot.Series, ...],
        *,
        mode: Literal[1, 2] | None = None,
        title: str | None = None,
        bands: tuple[uplot.Band, ...] | None = None,
        scales: dict[str, uplot.Scale] | None = None,
        axes: tuple[uplot.Axis, ...] | None = None,
        legend: uplot.Legend | None = None,
        cursor: uplot.Cursor | None = None,
        focus: uplot.Focus | None = None,
        aspect: float = 1.0,
        order: float | None = None,
        visible: bool = True,
    ) -> GuiUplotHandle:
        """Add a uPlot chart to the GUI for high-performance time-series visualization.

        uPlot is optimized for plotting large datasets with smooth pan/zoom interactions.
        All configuration options follow the standard uPlot API. For comprehensive
        documentation, see: https://github.com/leeoniya/uPlot/tree/1.6.32/docs

        .. note::
           Configuration types are exposed under the :mod:`viser.uplot` module for convenience:
           :class:`viser.uplot.Series`, :class:`viser.uplot.Scale`, :class:`viser.uplot.Axis`,
           :class:`viser.uplot.Band`, :class:`viser.uplot.Legend`, :class:`viser.uplot.Cursor`,
           and :class:`viser.uplot.Focus`. These are :py:class:`~typing.TypedDict` types;
           standard dictionaries can also be used.

        Args:
            data: Tuple of 1D numpy arrays containing chart data. The first array provides
                x-axis values, and subsequent arrays contain y-axis data for
                each series. All arrays must have identical length. Minimum 2
                arrays.
            series: Series configuration objects defining visual appearance and behavior.
                Must match the length of data tuple.
            mode: Chart layout mode. 1 = aligned (default) where all series share axes,
                2 = faceted where each series gets its own subplot panel.
            title: Chart title displayed at the top.
            bands: High/low range visualizations between data series. Useful for showing
                confidence intervals, error bounds, or min/max ranges. Each band connects
                two adjacent series indices.
            scales: Scale definitions controlling data-to-pixel mapping and axis ranges.
                Key features include auto-ranging, manual min/max, time-based scaling,
                and logarithmic distributions. Multiple scales enable dual-axis charts.
            axes: Axis configuration for labels, ticks, grids, and positioning.
                Controls which side axes appear (top/right/bottom/left), tick formatting,
                grid line styling, and spacing between tick marks.
            legend: Legend display options including positioning, styling, and custom
                value formatting functions for hover states.
            cursor: Interactive cursor behavior including hover proximity detection,
                drag-to-zoom, and crosshair appearance. Controls how users interact
                with the chart through mouse/touch.
            focus: Visual highlighting when hovering over series. Controls the alpha
                transparency of non-focused series to emphasize the active one.
            aspect: Width-to-height ratio for the chart display in the control panel.
                1.0 creates a square chart, values > 1.0 create wider charts.
            order: Display ordering relative to other GUI elements (lower values first).
            visible: Whether the chart is visible in the interface.

        Returns:
            A handle for programmatically updating chart properties and data.
        """

        # Validate data structure
        assert len(data) >= 2, (
            "data must have at least 2 arrays (x-data + at least one y-data series)"
        )
        assert all(isinstance(arr, np.ndarray) for arr in data), (
            "all data elements must be numpy arrays"
        )

        # Validate data dimensions and shapes
        for i, arr in enumerate(data):
            assert arr.ndim == 1, f"data[{i}] must be a 1D array, got shape {arr.shape}"

        # Check that all arrays have the same length
        lengths = [len(arr) for arr in data]
        if not all(length == lengths[0] for length in lengths):
            raise ValueError(
                f"All data arrays must have the same length. Got lengths: {lengths}"
            )

        # Validate series configuration
        assert len(series) > 0, "series must not be empty"
        if len(series) != len(data):
            raise ValueError(
                f"Length of series ({len(series)}) must match length of data ({len(data)}). "
                f"Each array in data needs a corresponding series configuration."
            )

        # Convert arrays to float64 as expected by GuiUplotProps
        data_float64 = tuple(arr.astype(np.float64) for arr in data)

        message = _messages.GuiUplotMessage(
            uuid=_make_uuid(),
            container_uuid=self._get_container_uuid(),
            props=_messages.GuiUplotProps(
                order=_apply_default_order(order),
                data=data_float64,
                mode=mode,
                title=title,
                series=series,
                bands=bands,
                scales=scales,
                axes=axes,
                legend=legend,
                cursor=cursor,
                focus=focus,
                aspect=aspect,
                visible=visible,
            ),
        )
        self._websock_interface.queue_message(message)

        return GuiUplotHandle(
            _GuiHandleState(
                message.uuid,
                self,
                value=None,
                props=message.props,
                parent_container_id=message.container_uuid,
            ),
        )

    @deprecated_positional_shim
    def add_button(
        self,
        label: str,
        *,
        disabled: bool = False,
        visible: bool = True,
        hint: str | None = None,
        color: LiteralColor | tuple[int, int, int] | None = None,
        icon: IconName | None = None,
        order: float | None = None,
    ) -> GuiButtonHandle:
        """Add a button to the GUI. The value of this input is set to `True` every time
        it is clicked; to detect clicks, we can manually set it back to `False`.

        Args:
            label: Label to display on the button.
            visible: Whether the button is visible.
            disabled: Whether the button is disabled.
            hint: Optional hint to display on hover.
            color: Optional color to use for the button.
            icon: Optional icon to display on the button.
            order: Optional ordering, smallest values will be displayed first.

        Returns:
            A handle that can be used to interact with the GUI element.
        """

        # Re-wrap the GUI handle with a button interface.
        uuid = _make_uuid()
        order = _apply_default_order(order)
        props = _messages.GuiButtonProps(
            order=order,
            label=label,
            hint=hint,
            color=color,
            _icon_html=None if icon is None else svg_from_icon(icon),
            _hold_callback_freqs=(),
            disabled=disabled,
            visible=visible,
        )
        message = _messages.GuiButtonMessage(
            value=False,
            uuid=uuid,
            container_uuid=self._get_container_uuid(),
            props=props,
        )

        # Send the message.
        self._websock_interface.queue_message(message)

        # Construct button-specific handle state with hold callback support.
        handle_state = _GuiButtonHandleState(
            props=props,
            gui_api=self,
            value=False,
            update_timestamp=time.time(),
            parent_container_id=self._get_container_uuid(),
            update_cb=[],
            is_button=True,
            sync_cb=None,
            uuid=uuid,
            hold_cbs_from_freq={},
        )

        return GuiButtonHandle(handle_state, _icon=icon)

    @deprecated_positional_shim
    def add_upload_button(
        self,
        label: str,
        *,
        disabled: bool = False,
        visible: bool = True,
        hint: str | None = None,
        color: LiteralColor | tuple[int, int, int] | None = None,
        icon: IconName | None = None,
        mime_type: str = "*/*",
        order: float | None = None,
    ) -> GuiUploadButtonHandle:
        """Add a button to the GUI. The value of this input is set to `True` every time
        it is clicked; to detect clicks, we can manually set it back to `False`.

        Args:
            label: Label to display on the button.
            visible: Whether the button is visible.
            disabled: Whether the button is disabled.
            hint: Optional hint to display on hover.
            color: Optional color to use for the button.
            icon: Optional icon to display on the button.
            mime_type: Optional MIME type to filter the files that can be uploaded.
            order: Optional ordering, smallest values will be displayed first.

        Returns:
            A handle that can be used to interact with the GUI element.
        """

        # Re-wrap the GUI handle with a button interface.
        uuid = _make_uuid()
        order = _apply_default_order(order)
        return GuiUploadButtonHandle(
            self._create_gui_input(
                value=UploadedFile("", b""),
                message=_messages.GuiUploadButtonMessage(
                    uuid=uuid,
                    container_uuid=self._get_container_uuid(),
                    props=_messages.GuiUploadButtonProps(
                        disabled=disabled,
                        visible=visible,
                        order=order,
                        label=label,
                        hint=hint,
                        color=color,
                        mime_type=mime_type,
                        _icon_html=None if icon is None else svg_from_icon(icon),
                    ),
                ),
                is_button=True,
            ),
            _icon=icon,
        )

    @deprecated_positional_shim
    def add_button_group(
        self,
        label: str,
        options: Sequence[str],
        *,
        disabled: bool = False,
        visible: bool = True,
        hint: str | None = None,
        order: float | None = None,
    ) -> GuiButtonGroupHandle:
        """Add a button group to the GUI.

        Args:
            label: Label to display on the button group.
            options: Sequence of options to display as buttons.
            disabled: Whether the button group is disabled.
            visible: Whether the button group is visible.
            hint: Optional hint to display on hover.
            order: Optional ordering, smallest values will be displayed first.

        Returns:
            A handle that can be used to interact with the GUI element.
        """
        value = options[0]
        uuid = _make_uuid()
        order = _apply_default_order(order)
        return GuiButtonGroupHandle(
            self._create_gui_input(
                value,
                message=_messages.GuiButtonGroupMessage(
                    value=value,
                    uuid=uuid,
                    container_uuid=self._get_container_uuid(),
                    props=_messages.GuiButtonGroupProps(
                        order=order,
                        label=label,
                        hint=hint,
                        options=tuple(options),
                        disabled=disabled,
                        visible=visible,
                    ),
                ),
                is_button=True,
            ),
        )

    @deprecated_positional_shim
    def add_checkbox(
        self,
        label: str,
        initial_value: bool,
        *,
        disabled: bool = False,
        visible: bool = True,
        hint: str | None = None,
        order: float | None = None,
    ) -> GuiCheckboxHandle:
        """Add a checkbox to the GUI.

        Args:
            label: Label to display on the checkbox.
            initial_value: Initial value of the checkbox.
            disabled: Whether the checkbox is disabled.
            visible: Whether the checkbox is visible.
            hint: Optional hint to display on hover.
            order: Optional ordering, smallest values will be displayed first.

        Returns:
            A handle that can be used to interact with the GUI element.
        """
        value = initial_value
        assert isinstance(value, bool)
        uuid = _make_uuid()
        order = _apply_default_order(order)
        return GuiCheckboxHandle(
            self._create_gui_input(
                value,
                message=_messages.GuiCheckboxMessage(
                    value=value,
                    uuid=uuid,
                    container_uuid=self._get_container_uuid(),
                    props=_messages.GuiCheckboxProps(
                        order=order,
                        label=label,
                        hint=hint,
                        disabled=disabled,
                        visible=visible,
                    ),
                ),
            )
        )

    @deprecated_positional_shim
    def add_text(
        self,
        label: str,
        initial_value: str,
        *,
        multiline: bool = False,
        disabled: bool = False,
        visible: bool = True,
        hint: str | None = None,
        order: float | None = None,
    ) -> GuiTextHandle:
        r"""Add a text input to the GUI.

        Args:
            label: Label to display on the text input.
            initial_value: Initial value of the text input.
            multiline: Whether the text input supports multiple lines, delimited with
                the \n character.
            disabled: Whether the text input is disabled.
            visible: Whether the text input is visible.
            hint: Optional hint to display on hover.
            order: Optional ordering, smallest values will be displayed first.

        Returns:
            A handle that can be used to interact with the GUI element.
        """
        value = initial_value
        assert isinstance(value, str)
        uuid = _make_uuid()
        order = _apply_default_order(order)
        return GuiTextHandle(
            self._create_gui_input(
                value,
                message=_messages.GuiTextMessage(
                    value=value,
                    uuid=uuid,
                    container_uuid=self._get_container_uuid(),
                    props=_messages.GuiTextProps(
                        order=order,
                        label=label,
                        hint=hint,
                        disabled=disabled,
                        visible=visible,
                        multiline=multiline,
                    ),
                ),
            )
        )

    @deprecated_positional_shim
    def add_number(
        self,
        label: str,
        initial_value: IntOrFloat,
        *,
        min: IntOrFloat | None = None,
        max: IntOrFloat | None = None,
        step: IntOrFloat | None = None,
        disabled: bool = False,
        visible: bool = True,
        hint: str | None = None,
        order: float | None = None,
    ) -> GuiNumberHandle[IntOrFloat]:
        """Add a number input to the GUI, with user-specifiable bound and precision parameters.

        Args:
            label: Label to display on the number input.
            initial_value: Initial value of the number input.
            min: Optional minimum value of the number input.
            max: Optional maximum value of the number input.
            step: Optional step size of the number input. Computed automatically if not
                specified.
            disabled: Whether the number input is disabled.
            visible: Whether the number input is visible.
            hint: Optional hint to display on hover.
            order: Optional ordering, smallest values will be displayed first.

        Returns:
            A handle that can be used to interact with the GUI element.
        """
        value = initial_value

        assert isinstance(value, (int, float))

        if step is None:
            # It's ok that `step` is always a float, even if the value is an integer,
            # because things all become `number` types after serialization.
            step = float(  # type: ignore
                np.min(
                    [
                        _compute_step(value),
                        _compute_step(min),
                        _compute_step(max),
                    ]
                )
            )

        assert step is not None

        uuid = _make_uuid()
        order = _apply_default_order(order)
        return GuiNumberHandle(
            self._create_gui_input(
                value,
                message=_messages.GuiNumberMessage(
                    value=value,
                    uuid=uuid,
                    container_uuid=self._get_container_uuid(),
                    props=_messages.GuiNumberProps(
                        order=order,
                        label=label,
                        hint=hint,
                        min=min,
                        max=max,
                        precision=_compute_precision_digits(step),
                        step=step,
                        disabled=disabled,
                        visible=visible,
                    ),
                ),
                is_button=False,
            )
        )

    @deprecated_positional_shim
    def add_vector2(
        self,
        label: str,
        initial_value: tuple[float, float] | np.ndarray,
        *,
        min: tuple[float, float] | np.ndarray | None = None,
        max: tuple[float, float] | np.ndarray | None = None,
        step: float | None = None,
        disabled: bool = False,
        visible: bool = True,
        hint: str | None = None,
        order: float | None = None,
    ) -> GuiVector2Handle:
        """Add a length-2 vector input to the GUI.

        Args:
            label: Label to display on the vector input.
            initial_value: Initial value of the vector input.
            min: Optional minimum value of the vector input.
            max: Optional maximum value of the vector input.
            step: Optional step size of the vector input. Computed automatically if not
            disabled: Whether the vector input is disabled.
            visible: Whether the vector input is visible.
            hint: Optional hint to display on hover.
            order: Optional ordering, smallest values will be displayed first.

        Returns:
            A handle that can be used to interact with the GUI element.
        """
        value = initial_value
        value = cast_vector(value, 2)
        min = cast_vector(min, 2) if min is not None else None
        max = cast_vector(max, 2) if max is not None else None
        uuid = _make_uuid()
        order = _apply_default_order(order)

        if step is None:
            possible_steps: list[float] = []
            possible_steps.extend([_compute_step(x) for x in value])
            if min is not None:
                possible_steps.extend([_compute_step(x) for x in min])
            if max is not None:
                possible_steps.extend([_compute_step(x) for x in max])
            step = float(np.min(possible_steps))

        return GuiVector2Handle(
            self._create_gui_input(
                value,
                message=_messages.GuiVector2Message(
                    value=value,
                    uuid=uuid,
                    container_uuid=self._get_container_uuid(),
                    props=_messages.GuiVector2Props(
                        order=order,
                        label=label,
                        hint=hint,
                        min=min,
                        max=max,
                        step=step,
                        precision=_compute_precision_digits(step),
                        disabled=disabled,
                        visible=visible,
                    ),
                ),
            )
        )

    @deprecated_positional_shim
    def add_vector3(
        self,
        label: str,
        initial_value: tuple[float, float, float] | np.ndarray,
        *,
        min: tuple[float, float, float] | np.ndarray | None = None,
        max: tuple[float, float, float] | np.ndarray | None = None,
        step: float | None = None,
        disabled: bool = False,
        visible: bool = True,
        hint: str | None = None,
        order: float | None = None,
    ) -> GuiVector3Handle:
        """Add a length-3 vector input to the GUI.

        Args:
            label: Label to display on the vector input.
            initial_value: Initial value of the vector input.
            min: Optional minimum value of the vector input.
            max: Optional maximum value of the vector input.
            step: Optional step size of the vector input. Computed automatically if not
            disabled: Whether the vector input is disabled.
            visible: Whether the vector input is visible.
            hint: Optional hint to display on hover.
            order: Optional ordering, smallest values will be displayed first.

        Returns:
            A handle that can be used to interact with the GUI element.
        """
        value = initial_value
        value = cast_vector(value, 3)
        min = cast_vector(min, 3) if min is not None else None
        max = cast_vector(max, 3) if max is not None else None
        uuid = _make_uuid()
        order = _apply_default_order(order)

        if step is None:
            possible_steps: list[float] = []
            possible_steps.extend([_compute_step(x) for x in value])
            if min is not None:
                possible_steps.extend([_compute_step(x) for x in min])
            if max is not None:
                possible_steps.extend([_compute_step(x) for x in max])
            step = float(np.min(possible_steps))

        return GuiVector3Handle(
            self._create_gui_input(
                value,
                message=_messages.GuiVector3Message(
                    value=value,
                    uuid=uuid,
                    container_uuid=self._get_container_uuid(),
                    props=_messages.GuiVector3Props(
                        order=order,
                        label=label,
                        hint=hint,
                        min=min,
                        max=max,
                        step=step,
                        precision=_compute_precision_digits(step),
                        disabled=disabled,
                        visible=visible,
                    ),
                ),
            )
        )

    # See add_dropdown for notes on overloads.
    @overload
    def add_dropdown(
        self,
        label: str,
        options: Sequence[TLiteralString],
        *,
        initial_value: TLiteralString | None = None,
        disabled: bool = False,
        visible: bool = True,
        hint: str | None = None,
        order: float | None = None,
    ) -> GuiDropdownHandle[TLiteralString]: ...

    @overload
    def add_dropdown(
        self,
        label: str,
        options: Sequence[TString],
        *,
        initial_value: TString | None = None,
        disabled: bool = False,
        visible: bool = True,
        hint: str | None = None,
        order: float | None = None,
    ) -> GuiDropdownHandle[TString]: ...

    @deprecated_positional_shim
    def add_dropdown(
        self,
        label: str,
        options: Sequence[TLiteralString] | Sequence[TString],
        *,
        initial_value: TLiteralString | TString | None = None,
        disabled: bool = False,
        visible: bool = True,
        hint: str | None = None,
        order: float | None = None,
    ) -> GuiDropdownHandle[Any]:  # Output type is specified in overloads.
        """Add a dropdown to the GUI.

        Args:
            label: Label to display on the dropdown.
            options: Sequence of options to display in the dropdown.
            initial_value: Initial value of the dropdown.
            disabled: Whether the dropdown is disabled.
            visible: Whether the dropdown is visible.
            hint: Optional hint to display on hover.
            order: Optional ordering, smallest values will be displayed first.

        Returns:
            A handle that can be used to interact with the GUI element.
        """
        value = initial_value
        if value is None:
            value = options[0]
        uuid = _make_uuid()
        order = _apply_default_order(order)
        return GuiDropdownHandle(
            self._create_gui_input(
                value,
                message=_messages.GuiDropdownMessage(
                    value=value,
                    uuid=uuid,
                    container_uuid=self._get_container_uuid(),
                    props=_messages.GuiDropdownProps(
                        order=order,
                        label=label,
                        hint=hint,
                        options=tuple(options),
                        disabled=disabled,
                        visible=visible,
                    ),
                ),
            ),
        )

    @deprecated_positional_shim
    def add_progress_bar(
        self,
        value: float,
        *,
        visible: bool = True,
        animated: bool = False,
        color: LiteralColor | tuple[int, int, int] | None = None,
        order: float | None = None,
    ) -> GuiProgressBarHandle:
        """Add a progress bar to the GUI.

        Args:
            value: Value of the progress bar. (0 - 100)
            visible: Whether the progress bar is visible.
            animated: Whether the progress bar is in a loading state (animated, striped).
            color: The color of the progress bar.
            order: Optional ordering, smallest values will be displayed first.

        Returns:
            A handle that can be used to interact with the GUI element.
        """
        assert value >= 0 and value <= 100
        message = _messages.GuiProgressBarMessage(
            value=value,
            uuid=_make_uuid(),
            container_uuid=self._get_container_uuid(),
            props=_messages.GuiProgressBarProps(
                order=_apply_default_order(order),
                animated=animated,
                color=color,
                visible=visible,
            ),
        )
        self._websock_interface.queue_message(message)
        handle = GuiProgressBarHandle(
            _GuiHandleState(
                message.uuid,
                self,
                value=value,
                props=message.props,
                parent_container_id=message.container_uuid,
            ),
        )
        return handle

    @deprecated_positional_shim
    def add_slider(
        self,
        label: str,
        min: IntOrFloat,
        max: IntOrFloat,
        step: IntOrFloat,
        initial_value: IntOrFloat,
        *,
        marks: tuple[IntOrFloat | tuple[IntOrFloat, str], ...] | None = None,
        disabled: bool = False,
        visible: bool = True,
        hint: str | None = None,
        order: float | None = None,
    ) -> GuiSliderHandle[IntOrFloat]:
        """Add a slider to the GUI. Types of the min, max, step, and initial value should match.

        Args:
            label: Label to display on the slider.
            min: Minimum value of the slider.
            max: Maximum value of the slider.
            step: Step size of the slider.
            initial_value: Initial value of the slider.
            marks: tuple of marks to display below the slider. Each mark should
                either be a numerical or a (number, label) tuple, where the
                label is provided as a string.
            disabled: Whether the slider is disabled.
            visible: Whether the slider is visible.
            hint: Optional hint to display on hover.
            order: Optional ordering, smallest values will be displayed first.

        Returns:
            A handle that can be used to interact with the GUI element.
        """
        value: IntOrFloat = initial_value
        assert max >= min
        step = builtins.min(step, max - min)
        assert max >= value >= min

        # GUI callbacks cast incoming values to match the type of the initial value. If
        # the min, max, or step is a float, we should cast to a float.
        #
        # This should also match what the IntOrFloat TypeVar resolves to.
        if type(value) is int and (
            type(min) is float or type(max) is float or type(step) is float
        ):
            value = float(value)  # type: ignore

        # TODO: as of 6/5/2023, this assert will break something in nerfstudio. (at
        # least LERF)
        #
        # assert type(min) == type(max) == type(step) == type(value)

        uuid = _make_uuid()
        order = _apply_default_order(order)
        return GuiSliderHandle(
            self._create_gui_input(
                value,
                message=_messages.GuiSliderMessage(
                    value=value,
                    uuid=uuid,
                    container_uuid=self._get_container_uuid(),
                    props=_messages.GuiSliderProps(
                        order=order,
                        label=label,
                        hint=hint,
                        min=min,
                        max=max,
                        step=step,
                        precision=_compute_precision_digits(step),
                        visible=visible,
                        disabled=disabled,
                        _marks=tuple(
                            GuiSliderMark(value=float(x[0]), label=x[1])
                            if isinstance(x, tuple)
                            else GuiSliderMark(value=x, label=None)
                            for x in marks
                        )
                        if marks is not None
                        else None,
                    ),
                ),
                is_button=False,
            )
        )

    @deprecated_positional_shim
    def add_multi_slider(
        self,
        label: str,
        min: IntOrFloat,
        max: IntOrFloat,
        step: IntOrFloat,
        initial_value: tuple[IntOrFloat, ...],
        *,
        min_range: IntOrFloat | None = None,
        fixed_endpoints: bool = False,
        marks: tuple[IntOrFloat | tuple[IntOrFloat, str], ...] | None = None,
        disabled: bool = False,
        visible: bool = True,
        hint: str | None = None,
        order: float | None = None,
    ) -> GuiMultiSliderHandle[IntOrFloat]:
        """Add a multi slider to the GUI. Types of the min, max, step, and initial value should match.

        Args:
            label: Label to display on the slider.
            min: Minimum value of the slider.
            max: Maximum value of the slider.
            step: Step size of the slider.
            initial_value: Initial values of the slider.
            min_range: Optional minimum difference between two values of the slider.
            fixed_endpoints: Whether the endpoints of the slider are fixed.
            marks: tuple of marks to display below the slider. Each mark should
                either be a numerical or a (number, label) tuple, where the
                label is provided as a string.
            disabled: Whether the slider is disabled.
            visible: Whether the slider is visible.
            hint: Optional hint to display on hover.
            order: Optional ordering, smallest values will be displayed first.

        Returns:
            A handle that can be used to interact with the GUI element.
        """
        assert max >= min
        step = builtins.min(step, max - min)
        assert all(max >= x >= min for x in initial_value)

        # GUI callbacks cast incoming values to match the type of the initial value. If
        # any of the arguments are floats, we should always use a float value.
        #
        # This should also match what the IntOrFloat TypeVar resolves to.
        if (
            type(min) is float
            or type(max) is float
            or type(step) is float
            or type(min_range) is float
        ):
            initial_value = tuple(float(x) for x in initial_value)  # type: ignore

        uuid = _make_uuid()
        order = _apply_default_order(order)
        return GuiMultiSliderHandle(
            self._create_gui_input(
                value=initial_value,
                message=_messages.GuiMultiSliderMessage(
                    value=initial_value,
                    uuid=uuid,
                    container_uuid=self._get_container_uuid(),
                    props=_messages.GuiMultiSliderProps(
                        order=order,
                        label=label,
                        hint=hint,
                        min=min,
                        min_range=min_range,
                        max=max,
                        step=step,
                        visible=visible,
                        disabled=disabled,
                        fixed_endpoints=fixed_endpoints,
                        precision=_compute_precision_digits(step),
                        _marks=tuple(
                            GuiSliderMark(value=float(x[0]), label=x[1])
                            if isinstance(x, tuple)
                            else GuiSliderMark(value=x, label=None)
                            for x in marks
                        )
                        if marks is not None
                        else None,
                    ),
                ),
                is_button=False,
            )
        )

    @deprecated_positional_shim
    def add_rgb(
        self,
        label: str,
        initial_value: tuple[int, int, int],
        *,
        disabled: bool = False,
        visible: bool = True,
        hint: str | None = None,
        order: float | None = None,
    ) -> GuiRgbHandle:
        """Add an RGB picker to the GUI. All values should be in [0, 255].

        Args:
            label: Label to display on the RGB picker.
            initial_value: Initial value of the RGB picker.
            disabled: Whether the RGB picker is disabled.
            visible: Whether the RGB picker is visible.
            hint: Optional hint to display on hover.
            order: Optional ordering, smallest values will be displayed first.

        Returns:
            A handle that can be used to interact with the GUI element.
        """

        value = initial_value
        uuid = _make_uuid()
        order = _apply_default_order(order)
        return GuiRgbHandle(
            self._create_gui_input(
                value,
                message=_messages.GuiRgbMessage(
                    value=value,
                    uuid=uuid,
                    container_uuid=self._get_container_uuid(),
                    props=_messages.GuiRgbProps(
                        order=order,
                        label=label,
                        hint=hint,
                        disabled=disabled,
                        visible=visible,
                    ),
                ),
            )
        )

    @deprecated_positional_shim
    def add_rgba(
        self,
        label: str,
        initial_value: tuple[int, int, int, int],
        *,
        disabled: bool = False,
        visible: bool = True,
        hint: str | None = None,
        order: float | None = None,
    ) -> GuiRgbaHandle:
        """Add an RGBA picker to the GUI. All values should be in [0, 255].

        Args:
            label: Label to display on the RGBA picker.
            initial_value: Initial value of the RGBA picker.
            disabled: Whether the RGBA picker is disabled.
            visible: Whether the RGBA picker is visible.
            hint: Optional hint to display on hover.
            order: Optional ordering, smallest values will be displayed first.

        Returns:
            A handle that can be used to interact with the GUI element.
        """
        value = initial_value
        uuid = _make_uuid()
        order = _apply_default_order(order)
        return GuiRgbaHandle(
            self._create_gui_input(
                value,
                message=_messages.GuiRgbaMessage(
                    value=value,
                    uuid=uuid,
                    container_uuid=self._get_container_uuid(),
                    props=_messages.GuiRgbaProps(
                        order=order,
                        label=label,
                        hint=hint,
                        disabled=disabled,
                        visible=visible,
                    ),
                ),
            )
        )

    class _GuiMessage(Protocol[GuiInputPropsType]):
        uuid: str
        props: GuiInputPropsType

    def _create_gui_input(
        self,
        value: T,
        message: _GuiMessage,
        is_button: bool = False,
    ) -> _GuiHandleState[T]:
        """Private helper for adding a simple GUI element."""

        # Send add GUI input message.
        assert isinstance(message, _messages.Message)
        self._websock_interface.queue_message(message)

        # Construct handle.
        handle_state = _GuiHandleState(
            props=message.props,
            gui_api=self,
            value=value,
            update_timestamp=time.time(),
            parent_container_id=self._get_container_uuid(),
            update_cb=[],
            is_button=is_button,
            sync_cb=None,
            uuid=message.uuid,
        )

        # For broadcasted GUI handles, we should synchronize all clients.
        # This will be a no-op for client handles.
        if not is_button:

            def sync_other_clients(
                client_id: ClientId, updates: dict[str, Any]
            ) -> None:
                message = _messages.GuiUpdateMessage(handle_state.uuid, updates)
                message.excluded_self_client = client_id
                self._websock_interface.queue_message(message)

            handle_state.sync_cb = sync_other_clients

        return handle_state



from __future__ import annotations

import abc
import asyncio
import atexit
import base64
import contextlib
import dataclasses
import gzip
import http
import logging
import mimetypes
import queue
import threading
import time
import webbrowser
from asyncio.events import AbstractEventLoop
from collections.abc import Coroutine
from pathlib import Path
from typing import Any, Callable, Generator, NewType, TypeVar

import msgspec.msgpack
import websockets.asyncio.server
import websockets.datastructures
import websockets.exceptions
import zstandard
from typing_extensions import Literal, assert_never, override
from websockets import Headers
from websockets.asyncio.server import ServerConnection
from websockets.http11 import Request, Response
from websockets.typing import Subprotocol

import viser  # Import for version checking

from ._async_message_buffer import AsyncMessageBuffer
from ._messages import Message


@dataclasses.dataclass
class _ClientHandleState:
    # Internal state for ClientConnection objects.
    # message_buffer: asyncio.Queue
    message_buffer: AsyncMessageBuffer
    event_loop: AbstractEventLoop


ClientId = NewType("ClientId", int)
TMessage = TypeVar("TMessage", bound=Message)


class StateSerializer:
    """Handle for serializing messages. In Viser, this is used to save the
    scene state so it can be shared/embedded in static webpages."""

    def __init__(
        self, handler: WebsockMessageHandler, filter: Callable[[Message], bool]
    ):
        self._handler = handler
        self._filter = filter
        self._time: float = 0.0
        self._messages: list[tuple[float, dict[str, Any]]] = []

    def _insert_message(self, message: Message) -> None:
        """Insert a message into the recorded file."""

        # Exclude messages that are filtered out. In Viser, this is typically
        # GUI messages.
        if not self._filter(message):
            return
        self._messages.append((self._time, message.as_serializable_dict()))

    def insert_sleep(self, duration: float) -> None:
        """Insert a sleep into the recorded file. This can be useful for
        dynamic 3D data."""
        assert self in self._handler._record_handles, "serialize() was already called!"
        self._time += duration

    def serialize(self) -> bytes:
        """Serialize saved messages. Should only be called once. Our convention
        is to write this binary format to a file with a ``.viser`` extension,
        for example via ``pathlib.Path("file.viser").write_bytes(...)``.

        Returns:
            The recording as bytes.
        """
        assert self in self._handler._record_handles, "serialize() was already called!"

        packed_bytes = msgspec.msgpack.encode(
            {
                "durationSeconds": self._time,
                "messages": self._messages,
                "viserVersion": viser.__version__,
            }
        )
        assert isinstance(packed_bytes, bytes)
        self._handler._record_handles.remove(self)
        # Use zstd for better compression ratio and speed.
        # Prepend 8-byte size header for decompressor.
        compressed = zstandard.ZstdCompressor(level=12).compress(packed_bytes)
        return len(packed_bytes).to_bytes(8, "little") + compressed

    def show(self, height: int = 400, dark_mode: bool = False) -> None:
        """Display the serialized scene in a Jupyter notebook or web browser.

        In Jupyter notebooks/labs, displays an inline IFrame. When running as a
        script, opens the visualization in the default web browser.

        See also :meth:`viser.ViserServer.show`.

        Args:
            height: Height of the embedded viewer in pixels.
            dark_mode: Use dark color scheme.
        """
        import html as html_module

        scene_bytes = self.serialize()
        scene_b64 = base64.b64encode(scene_bytes).decode("ascii")

        # Get client HTML and inject scene data as global variables.
        # The client reads from window.__VISER_EMBED_DATA__ (App.tsx).
        client_html_path = (
            Path(__file__).parent.parent / "client" / "build" / "index.html"
        )
        client_html = client_html_path.read_text()
        dark_mode_str = "true" if dark_mode else "false"
        inject_script = (
            f"<script>"
            f'window.__VISER_EMBED_DATA__="{scene_b64}";'
            f"window.__VISER_EMBED_CONFIG__={{darkMode:{dark_mode_str}}};"
            f"</script>"
        )
        modified_html = client_html.replace("</head>", f"{inject_script}</head>")

        # Display in IPython (Jupyter, Colab, myst-nb, etc.) using srcdoc.
        # This embeds the entire HTML inline, avoiding file serving issues.
        try:
            from IPython.core.getipython import get_ipython  # type: ignore

            ipython = get_ipython()
            if ipython is not None:
                from IPython.display import HTML, display  # type: ignore

                # Escape HTML for srcdoc attribute.
                escaped_html = html_module.escape(modified_html, quote=True)

                # Wrap in div to avoid IPython's "Consider using IFrame" warning.
                display(
                    HTML(
                        f'<div style="border:1px solid #ddd">'
                        f'<iframe srcdoc="{escaped_html}" '
                        f'width="100%" height="{height}" '
                        f'style="border:none;display:block"></iframe>'
                        f"</div>"
                    )
                )
                return
        except ImportError:
            pass

        # Fallback for scripts: write the modified HTML to a temp file.
        import tempfile

        with tempfile.NamedTemporaryFile(
            "w", suffix=".html", delete=False, encoding="utf-8"
        ) as f:
            f.write(modified_html)
            webbrowser.open("file://" + f.name)


class WebsockMessageHandler:
    """Mix-in for adding message handling to a class."""

    def __init__(self) -> None:
        self._incoming_handlers: dict[
            type[Message], list[Callable[[ClientId, Message], None | Coroutine]]
        ] = {}
        self._queued_messages: queue.Queue = queue.Queue()
        self._locked_thread_id = -1

        # List of active serializers recording messages.
        self._record_handles: list[StateSerializer] = []

    def get_message_serializer(
        self, filter: Callable[[Message], bool]
    ) -> StateSerializer:
        """Start recording messages that are sent. Sent messages will be
        serialized and can be used for playback."""
        serializer = StateSerializer(self, filter)
        self._record_handles.append(serializer)
        return serializer

    def register_handler(
        self,
        message_cls: type[TMessage],
        callback: Callable[[ClientId, TMessage], None | Coroutine],
    ) -> None:
        """Register a handler for a particular message type."""
        if message_cls not in self._incoming_handlers:
            self._incoming_handlers[message_cls] = []
        self._incoming_handlers[message_cls].append(callback)  # type: ignore

    def unregister_handler(
        self,
        message_cls: type[TMessage],
        callback: Callable[[ClientId, TMessage], None | Coroutine] | None = None,
    ):
        """Unregister a handler for a particular message type."""
        assert message_cls in self._incoming_handlers, (
            "Tried to unregister a handler that hasn't been registered."
        )
        if callback is None:
            self._incoming_handlers.pop(message_cls)
        else:
            self._incoming_handlers[message_cls].remove(callback)  # type: ignore

    async def _handle_incoming_message(
        self, client_id: ClientId, message: Message
    ) -> None:
        """Handle incoming messages."""
        if type(message) in self._incoming_handlers:
            for cb in self._incoming_handlers[type(message)]:
                if asyncio.iscoroutinefunction(cb):
                    await cb(client_id, message)
                else:
                    cb(client_id, message)

    @abc.abstractmethod
    def get_message_buffer(self) -> AsyncMessageBuffer: ...

    def queue_message(self, message: Message) -> None:
        """Wrapped method for sending messages."""
        for handle in self._record_handles:
            handle._insert_message(message)

        self.get_message_buffer().push(message)

    @contextlib.contextmanager
    def atomic(self) -> Generator[None, None, None]:
        """Returns a context where: all outgoing messages are grouped and applied by
        clients atomically.

        This should be treated as a soft constraint that's helpful for things
        like animations, or when we want position and orientation updates to
        happen synchronously.

        Returns:
            Context manager.
        """
        # If called multiple times in the same thread, we ignore inner calls.
        self.get_message_buffer().atomic_start()
        yield
        self.get_message_buffer().atomic_end()


class WebsockClientConnection(WebsockMessageHandler):
    """Handle for sending messages to and listening to messages from a single
    connected client."""

    def __init__(
        self,
        client_id: int,
        client_state: _ClientHandleState,
    ) -> None:
        self.client_id = client_id
        self._state = client_state
        super().__init__()

    @override
    def get_message_buffer(self) -> AsyncMessageBuffer:
        """Get client message buffer."""
        return self._state.message_buffer


class WebsockServer(WebsockMessageHandler):
    """Websocket server abstraction. Communicates asynchronously with client
    applications.

    By default, all messages are broadcasted to all connected clients.

    To send messages to an individual client, we can use `on_client_connect()` to
    retrieve client handles.

    Args:
        host: Host to bind server to.
        port: Port to bind server to.
        message_class: Base class for message types. Subclasses of the message type
            should have unique names. This argument is optional currently, but will be
            required in the future.
        http_server_root: Path to root for HTTP server.
        verbose: Toggle for print messages.
        client_api_version: Flag for backwards compatibility. 0 sends individual
            messages. 1 sends windowed messages.
    """

    def __init__(
        self,
        host: str,
        port: int,
        message_class: type[Message] = Message,
        http_server_root: Path | None = None,
        verbose: bool = True,
        client_api_version: Literal[0, 1] = 0,
    ):
        super().__init__()

        # Track connected clients.
        self._client_connect_cb: list[
            Callable[[WebsockClientConnection], None | Coroutine]
        ] = []
        self._client_disconnect_cb: list[
            Callable[[WebsockClientConnection], None | Coroutine]
        ] = []

        self._host = host
        self._port = port
        self._message_class = message_class
        self._http_server_root = http_server_root
        self._verbose = verbose
        self._client_api_version: Literal[0, 1] = client_api_version
        self._background_event_loop: asyncio.AbstractEventLoop | None = None

        self._stop_event: asyncio.Event | None = None

        self._client_state_from_id: dict[int, _ClientHandleState] = {}
        self._server_thread: threading.Thread | None = None

    def start(self) -> None:
        """Start the server."""

        # Start server thread.
        ready_sem = threading.Semaphore(value=1)
        ready_sem.acquire()
        self._server_thread = threading.Thread(
            target=lambda: self._background_worker(ready_sem),
            daemon=True,
        )
        self._server_thread.start()

        # Wait for ready signal from the background thread.
        ready_sem.acquire()

        # Exit the server thread when the main process exits. This would happen
        # automatically, but is nice to do explicitly to avoid some nanobind
        # reference leak warnings:
        # https://github.com/viser-project/viser/issues/518
        atexit.register(self.stop)

        # Broadcast buffer should be populated by the background worker.
        assert isinstance(self._broadcast_buffer, AsyncMessageBuffer)

    def stop(self) -> None:
        """Stop the server."""
        assert self._background_event_loop is not None
        assert self._stop_event is not None
        assert self._server_thread is not None

        # Unregister the atexit handler to prevent double-stop.
        atexit.unregister(self.stop)

        # Signal the background thread to stop.
        try:
            self._background_event_loop.call_soon_threadsafe(self._stop_event.set)
        except RuntimeError:
            # Event loop may already be closed during teardown.
            pass

        # Clean up the message buffers. This isn't really necessary, but helps
        # avoid "task destroyed" errors.
        self._broadcast_buffer.set_done()
        for client in self._client_state_from_id.values():
            client.message_buffer.set_done()

        # Wait for the server thread to finish.
        self._server_thread.join(timeout=0.1)

    def on_client_connect(
        self, cb: Callable[[WebsockClientConnection], None | Coroutine]
    ) -> None:
        """Attach a callback to run for newly connected clients."""
        self._client_connect_cb.append(cb)

    def on_client_disconnect(
        self, cb: Callable[[WebsockClientConnection], None | Coroutine]
    ) -> None:
        """Attach a callback to run when clients disconnect."""
        self._client_disconnect_cb.append(cb)

    @override
    def get_message_buffer(self) -> AsyncMessageBuffer:
        """Get the broadcast queue. Message will be sent to all clients."""
        return self._broadcast_buffer

    def flush(self) -> None:
        """Flush the outgoing message buffer for broadcasted messages. Any buffered
        messages will immediately be sent. (by default they are windowed)"""
        self._broadcast_buffer.flush()

    def flush_client(self, client_id: int) -> None:
        """Flush the outgoing message buffer for a particular client. Any buffered
        messages will immediately be sent. (by default they are windowed)"""
        # No-op if client is disconnected.
        client_state = self._client_state_from_id.get(client_id)
        if client_state is not None:
            client_state.message_buffer.flush()

    def _background_worker(self, ready_sem: threading.Semaphore) -> None:
        import rich

        host = self._host
        port = self._port
        message_class = self._message_class
        http_server_root = self._http_server_root

        # Need to make a new event loop for notebook compatbility.
        event_loop = asyncio.new_event_loop()
        asyncio.set_event_loop(event_loop)
        self._stop_event = asyncio.Event()
        self._background_event_loop = event_loop
        self._broadcast_buffer = AsyncMessageBuffer(
            event_loop, persistent_messages=True
        )

        count_lock = asyncio.Lock()
        connection_count = 0
        total_connections = 0

        async def ws_handler(
            connection: websockets.asyncio.server.ServerConnection,
        ) -> None:
            """Handler for websocket connections."""
            async with count_lock:
                nonlocal connection_count
                client_id = ClientId(connection_count)
                connection_count += 1

                nonlocal total_connections
                total_connections += 1

            # Version check to make sure Viser server/client match.
            if self._client_api_version == 1:
                import viser

                # Extract client version from the selected subprotocol.
                client_version_str = "unknown"
                if connection.subprotocol is not None:
                    if connection.subprotocol.startswith("viser-v"):
                        client_version_str = connection.subprotocol[7:].strip()

                if client_version_str != viser.__version__:
                    rich.print(
                        f"[bold red](viser)[/bold red] Version mismatch - connection rejected. "
                        f"Client: '{client_version_str}', Server: '{viser.__version__}'"
                    )
                    await connection.close(
                        1002,
                        f"Version mismatch. Client: {client_version_str}, Server: {viser.__version__}",
                    )
                    return  # Exit handler to prevent further processing.

            client_state = _ClientHandleState(
                AsyncMessageBuffer(event_loop, persistent_messages=False),
                event_loop,
            )
            client_connection = WebsockClientConnection(client_id, client_state)
            self._client_state_from_id[client_id] = client_state

            def handle_incoming(message: Message) -> None:
                event_loop.create_task(
                    self._handle_incoming_message(client_id, message)
                )
                event_loop.create_task(
                    client_connection._handle_incoming_message(client_id, message)
                )

            # New connection callbacks.
            for cb in self._client_connect_cb:
                if asyncio.iscoroutinefunction(cb):
                    await cb(client_connection)
                else:
                    cb(client_connection)

            if self._verbose:
                rich.print(
                    f"[bold](viser)[/bold] Connection opened ({client_id},"
                    f" {total_connections} total),"
                    f" {len(self._broadcast_buffer.message_from_id)} persistent"
                    " messages"
                )

            try:
                # For each client: infinite loop over producers (which send messages)
                # and consumers (which receive messages).
                await asyncio.gather(
                    _message_producer(
                        connection,
                        client_state.message_buffer,
                        client_id,
                        self._client_api_version,
                    ),
                    _message_producer(
                        connection,
                        self._broadcast_buffer,
                        client_id,
                        self._client_api_version,
                    ),
                    _message_consumer(connection, handle_incoming, message_class),
                )
            except (
                websockets.exceptions.ConnectionClosedOK,
                websockets.exceptions.ConnectionClosedError,
            ):
                # We use a sentinel value to signal that the client producer thread
                # should exit.
                #
                # This is partially cosmetic: it allows us to safely finish pending
                # queue get() tasks, which suppresses a "Task was destroyed but it is
                # pending" error.
                client_state.message_buffer.set_done()

                # Disconnection callbacks.
                for cb in self._client_disconnect_cb:
                    if asyncio.iscoroutinefunction(cb):
                        await cb(client_connection)
                    else:
                        cb(client_connection)

                # Cleanup.
                self._client_state_from_id.pop(client_id)
                total_connections -= 1
                if self._verbose:
                    rich.print(
                        f"[bold](viser)[/bold] Connection closed ({client_id},"
                        f" {total_connections} total)"
                    )

        # Host client on the same port as the websocket.
        file_cache: dict[Path, bytes] = {}
        file_cache_gzipped: dict[Path, bytes] = {}

        filter_added = False

        def viser_http_server(
            connection: ServerConnection,
            request: Request,
        ) -> Response | None:
            # <Hack>
            # Suppress errors for:
            # - https://github.com/python-websockets/websockets/issues/1513
            #    - (fixed in newer versions of websockets)
            # - https://github.com/python-websockets/websockets/issues/1606
            nonlocal filter_added
            if not filter_added:

                class NoHttpErrors(logging.Filter):
                    def filter(self, record):
                        return record.getMessage() not in (
                            "opening handshake failed",
                            "connection rejected (200 OK)",
                        )

                connection.logger.logger.addFilter(NoHttpErrors())  # type: ignore
                filter_added = True
            # </Hack>

            # Ignore websocket packets.
            if request.headers.get("Upgrade") == "websocket":
                return None

            # Strip out search params, get relative path.
            path = request.path
            path = path.partition("?")[0]
            relpath = str(Path(path).relative_to("/"))
            if relpath == ".":
                relpath = "index.html"
            assert http_server_root is not None

            source_path = http_server_root / relpath
            if not source_path.exists():
                return Response(http.HTTPStatus.NOT_FOUND, "NOT FOUND", Headers())

            use_gzip = "gzip" in request.headers.get("Accept-Encoding", "")

            # First, try some known MIME types. Using guess_type() can cause
            # problems for Javascript on some Windows machines.
            #
            # Some references:
            #     https://github.com/viser-project/viser/issues/256#issuecomment-2369684252
            #     https://bugs.python.org/issue43975
            #     https://github.com/golang/go/issues/32350#issuecomment-525111557
            #
            # We're assuming UTF-8, this is mostly reasonable but might want to revisit.
            mime_type = {
                ".css": "text/css; charset=utf-8",
                ".gif": "image/gif",
                ".htm": "text/html; charset=utf-8",
                ".html": "text/html; charset=utf-8",
                ".jpg": "image/jpeg",
                ".js": "application/javascript",
                ".wasm": "application/wasm",
                ".pdf": "application/pdf",
                ".png": "image/png",
                ".svg": "image/svg+xml",
                ".xml": "text/xml; charset=utf-8",
            }.get(Path(path).suffix.lower(), None)
            if mime_type is None:
                mime_type = mimetypes.guess_type(relpath)[0]
            if mime_type is None:
                mime_type = "application/octet-stream"

            if source_path not in file_cache:
                file_cache[source_path] = source_path.read_bytes()
            if use_gzip:
                if source_path not in file_cache_gzipped:
                    file_cache_gzipped[source_path] = gzip.compress(
                        file_cache[source_path]
                    )
                response_payload = file_cache_gzipped[source_path]
            else:
                response_payload = file_cache[source_path]

            response_headers = {
                "Content-Type": mime_type,
                "Content-Length": str(len(response_payload)),
                "Content-Encoding": "gzip" if use_gzip else "identity",
            }

            # Try to read + send over file.
            return Response(
                http.HTTPStatus.OK,
                "OK",
                websockets.datastructures.Headers(**response_headers),
                response_payload,
            )
            # return (http.HTTPStatus.OK, response_headers, response_payload)

        async def start_server() -> None:
            port_attempt = port
            for _ in range(1000):
                try:
                    async with websockets.asyncio.server.serve(
                        ws_handler,
                        host,
                        port_attempt,
                        # Increase ws message size limit to 50MB to allow large messages.
                        # (e.g. via client.get_render()).
                        max_size=50 * 1024 * 1024,
                        # Compression can be too slow for our use cases.
                        compression=None,
                        process_request=(
                            viser_http_server if http_server_root is not None else None
                        ),
                        # Accept connections with version-based protocol and extract version in handler.
                        subprotocols=None,
                        select_subprotocol=lambda _, subprotocols: next(
                            (
                                Subprotocol(p)
                                for p in subprotocols
                                if p.startswith("viser-v")
                            ),
                            None,
                        ),
                    ) as serve_future:
                        assert serve_future.server is not None
                        self._port = port_attempt
                        ready_sem.release()
                        assert self._stop_event is not None
                        await self._stop_event.wait()
                        return
                except OSError:  # Port not available.
                    port_attempt += 1
                    continue

        event_loop.run_until_complete(start_server())
        rich.print("[bold](viser)[/bold] Server stopped")

        # Clean up the event loop to prevent reference leaks
        event_loop.stop()
        event_loop.close()


async def _message_producer(
    websocket: ServerConnection,
    buffer: AsyncMessageBuffer,
    client_id: int,
    client_api_version: Literal[0, 1],
) -> None:
    """Infinite loop to broadcast windows of messages from a buffer."""
    window_generator = buffer.window_generator(client_id)
    zstd = zstandard.ZstdCompressor(level=1)
    while not buffer.done:
        try:
            outgoing = await window_generator.__anext__()
        except StopAsyncIteration:
            break

        if client_api_version == 1:
            # Encode the message structure.
            inner = msgspec.msgpack.encode(
                {
                    "messages": tuple(
                        message.as_serializable_dict() for message in outgoing
                    ),
                    "timestampSec": time.perf_counter(),
                }
            )
            # Compress and prepend size header (8 bytes, little-endian uint64).
            compressed = zstd.compress(inner)
            serialized = len(inner).to_bytes(8, "little") + compressed
            await websocket.send(serialized)
        elif client_api_version == 0:
            for msg in outgoing:
                serialized = msgspec.msgpack.encode(msg.as_serializable_dict())
                assert isinstance(serialized, bytes)
                await websocket.send(serialized)
        else:
            assert_never(client_api_version)


async def _message_consumer(
    websocket: ServerConnection,
    handle_message: Callable[[Message], None],
    message_class: type[Message],
) -> None:
    """Infinite loop waiting for and then handling incoming messages."""
    while True:
        raw = await websocket.recv()
        assert isinstance(raw, bytes)
        message = message_class.deserialize(raw)
        handle_message(message)


def error_print_wrapper(inner: Callable[[], Any]) -> Callable[[], None]:
    """Wrap a Callable to print error messages when they happen.

    This can be helpful for jobs submitted to ThreadPoolExecutor instances, which, by
    default, will suppress error messages until returned futures are awaited.
    """

    def wrapped() -> None:
        try:
            inner()
        except Exception as e:
            import traceback as tb

            tb.print_exception(type(e), e, e.__traceback__, limit=100)

    return wrapped



from __future__ import annotations

import dataclasses
from typing import Tuple

import numpy as np
import numpy.typing as npt
from typing_extensions import override

from . import _base, hints
from .utils import broadcast_leading_axes


@dataclasses.dataclass(frozen=True)
class SO2(
    _base.SOBase,
    matrix_dim=2,
    parameters_dim=2,
    tangent_dim=1,
    space_dim=2,
):
    """Special orthogonal group for 2D rotations. Broadcasting rules are the
    same as for numpy.

    Ported to numpy from `jaxlie.SO2`.

    Internal parameterization is `(cos, sin)`. Tangent parameterization is `(omega,)`.
    """

    # SO2-specific.

    unit_complex: npt.NDArray[np.floating]
    """Internal parameters. `(cos, sin)`. Shape should be `(*, 2)`."""

    @override
    def __repr__(self) -> str:
        unit_complex = np.round(self.unit_complex, 5)
        return f"{self.__class__.__name__}(unit_complex={unit_complex})"

    @staticmethod
    def from_radians(theta: hints.Scalar) -> SO2:
        """Construct a rotation object from a scalar angle."""
        cos = np.cos(theta)
        sin = np.sin(theta)
        return SO2(unit_complex=np.stack([cos, sin], axis=-1))

    def as_radians(self) -> npt.NDArray[np.floating]:
        """Compute a scalar angle from a rotation object."""
        radians = self.log()[..., 0]
        return radians

    # Factory.

    @classmethod
    @override
    def identity(
        cls, batch_axes: Tuple[int, ...] = (), dtype: npt.DTypeLike = np.float64
    ) -> SO2:
        return SO2(
            unit_complex=np.stack(
                [np.ones(batch_axes, dtype=dtype), np.zeros(batch_axes, dtype=dtype)],
                axis=-1,
            )
        )

    @classmethod
    @override
    def from_matrix(cls, matrix: npt.NDArray[np.floating]) -> SO2:
        assert matrix.shape[-2:] == (2, 2)
        return SO2(unit_complex=np.array(matrix[..., :, 0]))

    # Accessors.

    @override
    def as_matrix(self) -> npt.NDArray[np.floating]:
        cos_sin = self.unit_complex
        out = np.stack(
            [
                # [cos, -sin],
                cos_sin * np.array([1, -1], dtype=cos_sin.dtype),
                # [sin, cos],
                cos_sin[..., ::-1],
            ],
            axis=-2,
        )
        assert out.shape == (*self.get_batch_axes(), 2, 2)
        return out  # type: ignore

    @override
    def parameters(self) -> npt.NDArray[np.floating]:
        return self.unit_complex

    # Operations.

    @override
    def apply(self, target: npt.NDArray[np.floating]) -> npt.NDArray[np.floating]:
        assert target.shape[-1:] == (2,)
        self, target = broadcast_leading_axes((self, target))
        return np.einsum("...ij,...j->...i", self.as_matrix(), target)

    @override
    def multiply(self, other: SO2) -> SO2:  # type: ignore
        return SO2(
            unit_complex=np.einsum(
                "...ij,...j->...i", self.as_matrix(), other.unit_complex
            )
        )

    @classmethod
    @override
    def exp(cls, tangent: npt.NDArray[np.floating]) -> SO2:
        assert tangent.shape[-1] == 1
        cos = np.cos(tangent)
        sin = np.sin(tangent)
        return SO2(unit_complex=np.concatenate([cos, sin], axis=-1))

    @override
    def log(self) -> npt.NDArray[np.floating]:
        return np.arctan2(
            self.unit_complex[..., 1, None], self.unit_complex[..., 0, None]
        )

    @override
    def adjoint(self) -> npt.NDArray[np.floating]:
        return np.ones((*self.get_batch_axes(), 1, 1), dtype=self.unit_complex.dtype)

    @override
    def inverse(self) -> SO2:
        unit_complex = self.unit_complex.copy()
        unit_complex[..., 1] *= -1
        return SO2(unit_complex)

    @override
    def normalize(self) -> SO2:
        return SO2(
            unit_complex=self.unit_complex
            / np.linalg.norm(self.unit_complex, axis=-1, keepdims=True)
        )

    @classmethod
    @override
    def sample_uniform(
        cls,
        rng: np.random.Generator,
        batch_axes: Tuple[int, ...] = (),
        dtype: npt.DTypeLike = np.float64,
    ) -> SO2:
        out = SO2.from_radians(
            rng.uniform(0.0, 2.0 * np.pi, size=batch_axes).astype(dtype=dtype)
        )
        assert out.get_batch_axes() == batch_axes
        return out
